/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_scrn.c -- master for refresh, status bar, console, chat, notify, etc

#include "quakedef.h"

/*

  full screen console
  put up loading plaque
  blanked background with loading plaque
  blanked background with menu
  cinematics
  full screen image for quit and victory

  end of unit intermissions

*/


float		scr_con_current;	// aproaches scr_conlines at scr_conspeed
float		scr_conlines;		// 0.0 to 1.0 lines of console to display

qboolean scr_initialized;		// ready to draw

int			scr_draw_loading;

cvar_t		*scr_conspeed;
cvar_t		*scr_centertime;
cvar_t		*scr_showturtle;
cvar_t		*scr_showpause;
cvar_t		*scr_printspeed;

cvar_t		*scr_netgraph;
cvar_t		*scr_timegraph;
cvar_t		*scr_debuggraph;
cvar_t		*scr_graphheight;
cvar_t		*scr_graphscale;
cvar_t		*scr_graphshift;


char		crosshair_pic[MAX_QPATH];
int			crosshair_width, crosshair_height;

void SCR_TimeRefresh_f (void);
void SCR_Loading_f (void);
void SCR_ExecuteLayoutString (char *s);


/*
===============================================================================

BAR GRAPHS

===============================================================================
*/

/*
==============
CL_AddNetgraph

A new packet was just parsed
==============
*/
void CL_AddNetgraph (void)
{
	int		i;
	int		in;
	int		ping;

	// if using the debuggraph for something else, don't
	// add the net lines
	if (scr_debuggraph->value || scr_timegraph->value)
		return;

	for (i = 0; i < cls.netchan.dropped; i++)
		SCR_DebugGraph (30, 0x40);

	for (i = 0; i < cl.surpressCount; i++)
		SCR_DebugGraph (30, 0xdf);

	// see what the latency was on this packet
	in = cls.netchan.incoming_acknowledged & (CMD_BACKUP - 1);
	ping = cls.realtime - cl.cmd_time[in];
	ping /= 30;

	if (ping > 30)
		ping = 30;

	SCR_DebugGraph (ping, 0xd0);
}


typedef struct
{
	float	value;
	int		color;
} graphsamp_t;

static	int			current;
static	graphsamp_t	values[1024];

/*
==============
SCR_DebugGraph
==============
*/
void SCR_DebugGraph (float value, int color)
{
	values[current & 1023].value = value;
	values[current & 1023].color = color;
	current++;
}

/*
==============
SCR_DrawDebugGraph
==============
*/
void SCR_DrawDebugGraph (void)
{
	int		a, x, y, w, i, h;
	float	v;
	int		color;

	// draw the graph
	w = viddef.width;

	x = 0;
	y = viddef.height;

	re->DrawFill (x, y - scr_graphheight->value, w, scr_graphheight->value, 8);

	for (a = 0; a < w; a++)
	{
		i = (current - 1 - a + 1024) & 1023;
		v = values[i].value;
		color = values[i].color;
		v = v * scr_graphscale->value + scr_graphshift->value;

		if (v < 0)
			v += scr_graphheight->value * (1 + (int) (-v / scr_graphheight->value));

		h = (int) v % (int) scr_graphheight->value;
		re->DrawFill (x + w - 1 - a, y - h, 1, h, color);
	}
}

/*
===============================================================================

CENTER PRINTING

===============================================================================
*/

char		scr_centerstring[1024];
float		scr_centertime_start;	// for slow victory printing
float		scr_centertime_off;
int			scr_center_lines;
int			scr_erase_center;

/*
==============
SCR_CenterPrint

Called for important messages that should stay in the center of the screen
for a few moments
==============
*/
void SCR_CenterPrint (char *str)
{
	char	*s;
	char	line[64];
	int		i, j, l;

	Q_strncpyz (scr_centerstring, str, sizeof (scr_centerstring));
	scr_centertime_off = scr_centertime->value;
	scr_centertime_start = cl.time;

	// count the number of lines for centering
	scr_center_lines = 1;
	s = str;

	while (*s)
	{
		if (*s == '\n')
			scr_center_lines++;

		s++;
	}

	// echo it to the console
	Com_Printf ("\n\n\35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n\n");

	s = str;

	do
	{
		// scan the width of the line
		for (l = 0; l < 40; l++)
		{
			if (s[l] == '\n' || !s[l])
				break;
		}

		for (i = 0; i < (40 - l) / 2; i++)
			line[i] = ' ';

		for (j = 0; j < l; j++)
		{
			line[i++] = s[j];
		}

		line[i] = '\n';
		line[i + 1] = 0;

		Com_Printf ("%s", line);

		while (*s && *s != '\n')
			s++;

		if (!*s)
			break;

		s++;		// skip the \n
	} while (1);

	Com_Printf ("\n\n\35\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\36\37\n\n");
	Con_ClearNotify ();
}


void SCR_DrawCenterString (void)
{
	char	*start;
	int		l;
	int		j;
	int		x, y;

	scr_erase_center = 0;
	start = scr_centerstring;

	y = viddef.height * 0.35 - scr_center_lines * 4;

	do
	{
		// scan the width of the line
		for (l = 0; l < 40; l++)
		{
			if (start[l] == '\n' || !start[l])
				break;
		}

		x = (viddef.width - l * 8) / 2;

		re->DrawString (x, y, start, l, 0);

		y += 8;

		while (*start && *start != '\n')
			start++;

		if (!*start)
			break;

		start++;		// skip the \n
	} while (1);
}

void SCR_CheckDrawCenterString (void)
{
	scr_centertime_off -= cls.frametime;

	if (scr_centertime_off <= 0)
		return;

	SCR_DrawCenterString ();
}

//=============================================================================


/*
=================
SCR_Sky_f

Set a specific sky and rotation speed
=================
*/
void SCR_Sky_f (void)
{
	float	rotate;
	vec3_t	axis;

	if (Cmd_Argc () < 2)
	{
		Com_Printf ("Usage: %s <basename> <rotate> <axis x y z>\n", Cmd_Argv (0));
		return;
	}

	if (Cmd_Argc () > 2)
		rotate = atof (Cmd_Argv (2));
	else rotate = 0;

	if (Cmd_Argc () == 6)
	{
		axis[0] = atof (Cmd_Argv (3));
		axis[1] = atof (Cmd_Argv (4));
		axis[2] = atof (Cmd_Argv (5));
	}
	else
	{
		axis[0] = 0;
		axis[1] = 0;
		axis[2] = 1;
	}

	re->SetSky (Cmd_Argv (1), rotate, axis);
}

//============================================================================

/*
==================
SCR_Init
==================
*/
void SCR_Init (void)
{
	scr_conspeed = Cvar_Get ("scr_conspeed", "3", 0);
	scr_showturtle = Cvar_Get ("scr_showturtle", "0", 0);
	scr_showpause = Cvar_Get ("scr_showpause", "1", 0);
	scr_centertime = Cvar_Get ("scr_centertime", "2.5", 0);
	scr_printspeed = Cvar_Get ("scr_printspeed", "8", 0);
	scr_netgraph = Cvar_Get ("netgraph", "0", 0);
	scr_timegraph = Cvar_Get ("timegraph", "0", 0);
	scr_debuggraph = Cvar_Get ("debuggraph", "0", 0);
	scr_graphheight = Cvar_Get ("graphheight", "32", 0);
	scr_graphscale = Cvar_Get ("graphscale", "1", 0);
	scr_graphshift = Cvar_Get ("graphshift", "0", 0);

	// register our commands
	Cmd_AddCommand ("timerefresh", SCR_TimeRefresh_f);
	Cmd_AddCommand ("loading", SCR_Loading_f);
	Cmd_AddCommand ("sky", SCR_Sky_f);
	Cmd_AddCommand ("loadsky", SCR_Sky_f);

	scr_initialized = true;
}


/*
==============
SCR_DrawNet
==============
*/
void SCR_DrawNet (void)
{
	if (cls.netchan.outgoing_sequence - cls.netchan.incoming_acknowledged < CMD_BACKUP - 1)
		return;

	re->DrawPic (64, 0, "net");
}

/*
==============
SCR_DrawPause
==============
*/
void SCR_DrawPause (void)
{
	int	w, h;

	if (!scr_showpause->value)		// turn off for screenshots
		return;

	if (!cl_paused->value)
		return;

	// i suspect this was offset differently so that it would appear when the console was down but a comment would have been nice
	if (re->DrawGetPicSize (&w, &h, "pause"))
		re->DrawPic ((viddef.width - w) / 2, (viddef.height - h) / 2, "pause");
}

/*
==============
SCR_DrawLoading
==============
*/
void SCR_DrawLoading (void)
{
	int	w, h;

	if (!scr_draw_loading)
		return;

	// huh?  surely should be true???
	// or is this a guard against calling recursively???
	scr_draw_loading = false;

	if (re->DrawGetPicSize (&w, &h, "loading"))
		re->DrawPic ((viddef.width - w) / 2, (viddef.height - h) / 2, "loading");
}

//=============================================================================

/*
==================
SCR_RunConsole

Scroll it up or down
==================
*/
void SCR_RunConsole (void)
{
	static int con_oldtime = -1;
	float con_frametime;

	// check for first call
	if (con_oldtime < 0) con_oldtime = cls.realtime;

	// get correct frametime
	con_frametime = (float) (cls.realtime - con_oldtime) * 0.001;
	con_oldtime = cls.realtime;

	// decide on the height of the console
	if (cls.key_dest == key_console)
		scr_conlines = 0.5;		// half screen
	else scr_conlines = 0;		// none visible

	if (scr_conlines < scr_con_current)
	{
		scr_con_current -= scr_conspeed->value * (viddef.height / 200.0f) * con_frametime;

		if (scr_conlines > scr_con_current)
			scr_con_current = scr_conlines;
	}
	else if (scr_conlines > scr_con_current)
	{
		scr_con_current += scr_conspeed->value * (viddef.height / 200.0f) * con_frametime;

		if (scr_conlines < scr_con_current)
			scr_con_current = scr_conlines;
	}
}


/*
==================
SCR_DrawConsole
==================
*/
void SCR_DrawConsole (void)
{
	Con_CheckResize ();

	if (cls.state == ca_disconnected || cls.state == ca_connecting)
	{
		// forced full screen console
		Con_DrawConsole (1.0, 1.0);
		return;
	}

	if (cls.state != ca_active || !cl.refresh_prepped)
	{
		// connected, but can't render
		re->Clear (0);
		Con_DrawConsole (0.5, 1.0);
		return;
	}

	if (scr_con_current)
	{
		Con_DrawConsole (scr_con_current, scr_con_current * 1.25f);
	}
	else
	{
		if (cls.key_dest == key_game || cls.key_dest == key_message)
			Con_DrawNotify ();	// only draw notify in game
	}
}

//=============================================================================

/*
================
SCR_BeginLoadingPlaque
================
*/
void SCR_BeginLoadingPlaque (void)
{
	S_StopAllSounds ();
	cl.sound_prepped = false;		// don't play ambients

	if (cls.disable_screen)
		return;

	if (developer->value)
		return;

	if (cls.state == ca_disconnected)
		return;	// if at console, don't bring up the plaque

	if (cls.key_dest == key_console)
		return;

	if (cl.cinematictime > 0)
		scr_draw_loading = 2;	// clear to black first
	else scr_draw_loading = 1;

	SCR_UpdateScreen ();
	cls.disable_screen = Sys_Milliseconds ();
	cls.disable_servercount = cl.servercount;
}

/*
================
SCR_EndLoadingPlaque
================
*/
void SCR_EndLoadingPlaque (void)
{
	cls.disable_screen = 0;
	Con_ClearNotify ();
}


/*
================
SCR_Loading_f
================
*/
void SCR_Loading_f (void)
{
	SCR_BeginLoadingPlaque ();
}

/*
================
SCR_TimeRefresh_f
================
*/
int entitycmpfnc (const entity_t *a, const entity_t *b)
{
	// all other models are sorted by model then skin
	if (a->model == b->model)
		return ((int) a->skin - (int) b->skin);
	else return ((int) a->model - (int) b->model);
}

void SCR_TimeRefresh_f (void)
{
	int	num;

	if (cls.state != ca_active)
		return;

	// check argc and run a number of iterations based on that
	if (Cmd_Argc () > 1)
		num = atoi (Cmd_Argv (1));
	else num = 128;

	// bound sensibly
	if (num < 10) num = 10;
	if (num > 1000) num = 1000;

	int start = Sys_Milliseconds ();

	for (int i = 0; i < num; i++)
	{
		// fixme - starting at current viewangle, do the spin from there
		cl.refdef.viewangles[1] = ((float) i * 360.0f) / (float) num;

		if (re->BeginFrame ())
		{
			re->RenderFrame (&cl.refdef);
			re->EndFrame ();
		}
	}

	int stop = Sys_Milliseconds ();
	float time = (float) (stop - start) / 1000.0f;

	Com_Printf ("%f seconds (%f fps)\n", time, num / time);
}


//===============================================================


/*
===============
SCR_TouchPics

Allows rendering code to cache all needed sbar graphics
===============
*/
void SCR_TouchPics (void)
{
	if (crosshair->value)
	{
		if (crosshair->value > 3 || crosshair->value < 0)
			crosshair->value = 3;

		Com_sprintf (crosshair_pic, sizeof (crosshair_pic), "ch%i", (int) (crosshair->value));

		if (!re->DrawGetPicSize (&crosshair_width, &crosshair_height, crosshair_pic))
			crosshair_pic[0] = 0;
	}
}

/*
================
SCR_DrawStats

The status bar is a small layout program that
is based on the stats array
================
*/
void SCR_DrawStats (void)
{
	SCR_ExecuteLayoutString (cl.configstrings[CS_STATUSBAR]);
}


/*
================
SCR_DrawLayout

================
*/
#define	STAT_LAYOUTS		13

void SCR_DrawLayout (void)
{
	if (!cl.frame.playerstate.stats[STAT_LAYOUTS])
		return;

	SCR_ExecuteLayoutString (cl.layout);
}

//=======================================================

/*
==================
SCR_UpdateScreen

This is called every frame, and can also be called explicitly to flush
text to the screen.
==================
*/
void CL_DrawFPS (void);

void SCR_UpdateScreen (void)
{
	// if the screen is disabled (loading plaque is up, or vid mode changing)
	// do nothing at all
	if (cls.disable_screen)
	{
		if (Sys_Milliseconds () - cls.disable_screen > 120000)
		{
			cls.disable_screen = 0;
			Com_Printf ("Loading plaque timed out.\n");
		}

		return;
	}

	// not initialized yet
	if (!scr_initialized || !con.initialized) return;

	// d3d device failed to create or is currently lost
	if (!re->BeginFrame ()) return;

	if (scr_draw_loading == 2)
	{
		//  loading plaque over black screen
		int		w, h;

		re->Set2DMode ();
		re->SetCinematicPalette (NULL);
		scr_draw_loading = 0;
		re->Clear (0);	// this was never done... (it would have been re->DrawFill in the original)

		if (re->DrawGetPicSize (&w, &h, "loading"))
			re->DrawPic ((viddef.width - w) / 2, (viddef.height - h) / 2, "loading");
	}
	else if (cl.cinematictime > 0)
	{
		re->Set2DMode ();
		re->Clear (0);

		// if a cinematic is supposed to be running, handle menus and console specially
		if (cls.key_dest == key_menu)
			M_Draw ();
		else if (cls.key_dest == key_console)
			SCR_DrawConsole ();
		else SCR_DrawCinematic ();
	}
	else
	{
		// do 3D refresh drawing, and then update the screen
		V_RenderView ();

		// 2D stuff goes here
		re->Set2DMode ();

		CL_DrawFPS ();

		SCR_DrawStats ();

		if (cl.frame.playerstate.stats[STAT_LAYOUTS] & 1) SCR_DrawLayout ();
		if (cl.frame.playerstate.stats[STAT_LAYOUTS] & 2) CL_DrawInventory ();

		SCR_DrawNet ();
		SCR_CheckDrawCenterString ();

		if (scr_timegraph->value)
			SCR_DebugGraph (cls.frametime * 300, 0);

		if (scr_debuggraph->value || scr_timegraph->value || scr_netgraph->value)
			SCR_DrawDebugGraph ();

		SCR_DrawPause ();

		SCR_DrawConsole ();

		M_Draw ();

		SCR_DrawLoading ();
	}

	re->EndFrame ();
}
