/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// cl_view.c -- player rendering positioning

#include "quakedef.h"


cvar_t		*crosshair;
cvar_t		*cl_testparticles;
cvar_t		*cl_testentities;
cvar_t		*cl_testlights;
cvar_t		*cl_testblend;
cvar_t		*cl_fov;
cvar_t		*cl_stats;


int			r_numdlights;
dlight_t	r_dlights[MAX_DLIGHTS];

int			r_numentities;
entity_t	r_entities[MAX_ENTITIES];

int			r_numparticles;
particle_t	r_particles[MAX_PARTICLES];

float		r_lightstyles[MAX_LIGHTSTYLES];

char cl_weaponmodels[MAX_CLIENTWEAPONMODELS][MAX_QPATH];
int num_cl_weaponmodels;

/*
====================
V_ClearScene

Specifies the model that will be used as the world
====================
*/
void V_ClearScene (void)
{
	r_numdlights = 0;
	r_numentities = 0;
	r_numparticles = 0;
}


/*
=====================
V_AddEntity

=====================
*/
void V_AddEntity (entity_t *ent)
{
	if (r_numentities >= MAX_ENTITIES)
		return;

	// and add it
	r_entities[r_numentities] = *ent;

	// clear any render fx not defined by the engine in case mods misuse this for their own internal crap.
	// this allows us to define our own in-engine effects above RF_USE_DISGUISE and not have them stomped on.
	// it's safe to do this now without stomping on mods ourselves because the entity has been copied.
	// using 0x79FFF instead of 0x7FFFF because there's a gap between (1 << 12) and (1 << 15)
	r_entities[r_numentities].flags &= 0x79FFF;

	// go to the next entity
	r_numentities++;
}


/*
=====================
V_AddParticle

=====================
*/
void V_AddParticle (vec3_t org, int color, float alpha)
{
	if (r_numparticles < MAX_PARTICLES)
	{
		particle_t	*p = &r_particles[r_numparticles];

		VectorCopy (org, p->origin);
		p->color = re->palette[color & 255];
		p->alpha = alpha;

		r_numparticles++;
	}
}


/*
=====================
V_AddLight

=====================
*/
void V_AddLight (vec3_t org, float intensity, float r, float g, float b)
{
	if (r_numdlights >= MAX_DLIGHTS) return;

	dlight_t *dl = &r_dlights[r_numdlights];

	VectorCopy (org, dl->origin);

	dl->intensity = intensity;

	// check for negative lights
	if (r < 0 || g < 0 || b < 0)
	{
		dl->color[0] = -r;
		dl->color[1] = -g;
		dl->color[2] = -b;
		dl->negative = true;
	}
	else
	{
		dl->color[0] = r;
		dl->color[1] = g;
		dl->color[2] = b;
		dl->negative = false;
	}

	// normalize lighting to 1, 1, 1 scale
	Vector3Normalize (dl->color);

	r_numdlights++;
}


void V_AddDLight (vec3_t org, float intensity, int r, int g, int b)
{
	V_AddLight (org, intensity, (float) r / 256.0f, (float) g / 256.0f, (float) b / 256.0f);
}


/*
=====================
V_AddLightStyle

=====================
*/
void V_AddLightStyle (int style, float intensity)
{
	if (style < 0 || style > MAX_LIGHTSTYLES)
		Com_Error (ERR_DROP, "Bad light style %i", style);

	r_lightstyles[style] = intensity;
}


/*
================
V_TestParticles

If cl_testparticles is set, create 4096 particles in the view
================
*/
void V_TestParticles (void)
{
	particle_t	*p;
	int			i, j;
	float		d, r, u;

	r_numparticles = cl_testparticles->value * 1024;

	if (r_numparticles < 1) r_numparticles = 0;
	if (r_numparticles > MAX_PARTICLES) r_numparticles = MAX_PARTICLES;

	for (i = 0; i < r_numparticles; i++)
	{
		d = i * 0.25;
		r = 4 * ((i & 7) - 3.5);
		u = 4 * (((i >> 3) & 7) - 3.5);
		p = &r_particles[i];

		for (j = 0; j < 3; j++)
			p->origin[j] = cl.refdef.vieworigin[j] + cl.v_forward[j] * d + cl.v_right[j] * r + cl.v_up[j] * u;

		p->color = 8;
		p->alpha = 1;
	}
}

/*
================
V_TestEntities

If cl_testentities is set, create 32 player models
================
*/
void V_TestEntities (void)
{
	int			i, j;
	float		f, r;
	entity_t	*ent;

	r_numentities = 32;
	memset (r_entities, 0, sizeof (r_entities));

	for (i = 0; i < r_numentities; i++)
	{
		ent = &r_entities[i];

		r = 64 * ((i % 4) - 1.5);
		f = 64 * (i / 4) + 128;

		for (j = 0; j < 3; j++)
			ent->currorigin[j] = cl.refdef.vieworigin[j] + cl.v_forward[j] * f + cl.v_right[j] * r;

		ent->model = cl.baseclientinfo.model;
		ent->skin = cl.baseclientinfo.skin;
	}
}

/*
================
V_TestLights

If cl_testlights is set, create 32 lights models
================
*/
void V_TestLights (void)
{
	r_numdlights = cl_testlights->value;

	if (r_numdlights < 1) r_numdlights = 0;
	if (r_numdlights > MAX_DLIGHTS) r_numdlights = MAX_DLIGHTS;

	memset (r_dlights, 0, sizeof (r_dlights));

	for (int i = 0; i < r_numdlights; i++)
	{
		dlight_t *dl = &r_dlights[i];

		float r = 64 * ((i % 4) - 1.5);
		float f = 64 * (i / 4) + 128;

		for (int j = 0; j < 3; j++)
			dl->origin[j] = cl.refdef.vieworigin[j] + cl.v_forward[j] * f + cl.v_right[j] * r;

		dl->color[0] = ((i % 6) + 1) & 1;
		dl->color[1] = (((i % 6) + 1) & 2) >> 1;
		dl->color[2] = (((i % 6) + 1) & 4) >> 2;
		dl->intensity = 200;
	}
}

//===================================================================

/*
=================
CL_PrepRefresh

Call before entering a new level, or after changing dlls
=================
*/
void CL_PrepRefresh (void)
{
	char		mapname[32];
	int			i;
	char		name[MAX_QPATH];
	float		rotate;
	vec3_t		axis;

	if (!cl.configstrings[CS_MODELS + 1][0])
		return;		// no map loaded

	// let the render dll load the map
	strcpy (mapname, cl.configstrings[CS_MODELS + 1] + 5);	// skip "maps/"
	mapname[strlen (mapname) - 4] = 0;		// cut off ".bsp"

	// register models, pics, and skins
	Com_Printf ("Map: %s\r", mapname);
	SCR_UpdateScreen ();
	re->BeginRegistration (mapname);
	Com_Printf ("                                     \r");

	// precache status bar pics
	Com_Printf ("pics\r");
	SCR_UpdateScreen ();
	SCR_TouchPics ();
	Com_Printf ("                                     \r");

	CL_RegisterTEntModels ();

	num_cl_weaponmodels = 1;
	strcpy (cl_weaponmodels[0], "weapon.md2");

	for (i = 1; i < MAX_MODELS && cl.configstrings[CS_MODELS + i][0]; i++)
	{
		strcpy (name, cl.configstrings[CS_MODELS + i]);
		name[37] = 0;	// never go beyond one line

		if (name[0] != '*')
			Com_Printf ("%s\r", name);

		SCR_UpdateScreen ();
		Sys_SendKeyEvents ();	// pump message loop

		if (name[0] == '#')
		{
			// special player weapon model
			if (num_cl_weaponmodels < MAX_CLIENTWEAPONMODELS)
			{
				Q_strncpyz (cl_weaponmodels[num_cl_weaponmodels], cl.configstrings[CS_MODELS + i] + 1, sizeof (cl_weaponmodels[num_cl_weaponmodels]));
				num_cl_weaponmodels++;
			}
		}
		else
		{
			cl.model_draw[i] = re->RegisterModel (cl.configstrings[CS_MODELS + i]);

			if (name[0] == '*')
				cl.model_clip[i] = CM_InlineModel (cl.configstrings[CS_MODELS + i]);
			else cl.model_clip[i] = NULL;
		}

		if (name[0] != '*')
			Com_Printf ("                                     \r");
	}

	Com_Printf ("images\r", i);
	SCR_UpdateScreen ();

	for (i = 1; i < MAX_IMAGES && cl.configstrings[CS_IMAGES + i][0]; i++)
	{
		cl.image_precache[i] = re->RegisterPic (cl.configstrings[CS_IMAGES + i]);
		Sys_SendKeyEvents ();	// pump message loop
	}

	Com_Printf ("                                     \r");

	for (i = 0; i < MAX_CLIENTS; i++)
	{
		if (!cl.configstrings[CS_PLAYERSKINS + i][0])
			continue;

		Com_Printf ("client %i\r", i);
		SCR_UpdateScreen ();
		Sys_SendKeyEvents ();	// pump message loop
		CL_ParseClientinfo (i);
		Com_Printf ("                                     \r");
	}

	CL_LoadClientinfo (&cl.baseclientinfo, "unnamed\\male/grunt");

	// set sky textures and speed
	Com_Printf ("sky\r", i);
	SCR_UpdateScreen ();
	rotate = atof (cl.configstrings[CS_SKYROTATE]);
	sscanf (cl.configstrings[CS_SKYAXIS], "%f %f %f", &axis[0], &axis[1], &axis[2]);
	re->SetSky (cl.configstrings[CS_SKY], rotate, axis);
	Com_Printf ("                                     \r");

	// the renderer can now free unneeded stuff
	re->EndRegistration ();

	// clear any lines of console text
	Con_ClearNotify ();

	SCR_UpdateScreen ();
	cl.refresh_prepped = true;
	cl.force_refdef = true;	// make sure we have a valid refdef
}


/*
====================
CalcFov
====================
*/
float SCR_CalcFovX (float fov_y, float width, float height)
{
	// bound, don't crash
	if (fov_y < 1) fov_y = 1;
	if (fov_y > 179) fov_y = 179;

	return (atan (width / (height / tan ((fov_y * M_PI) / 360.0f))) * 360.0f) / M_PI;
}


float SCR_CalcFovY (float fov_x, float width, float height)
{
	// bound, don't crash
	if (fov_x < 1) fov_x = 1;
	if (fov_x > 179) fov_x = 179;

	return (atan (height / (width / tan ((fov_x * M_PI) / 360.0f))) * 360.0f) / M_PI;
}


void SCR_SetFOV (refdef_t *rd, float fovvar, int width, int height)
{
	float aspect = (float) height / (float) width;

	// set up relative to a baseline aspect of 640x480
#define BASELINE_W	640.0f
#define BASELINE_H	480.0f

	// http://www.gamedev.net/topic/431111-perspective-math-calculating-horisontal-fov-from-vertical/
	// horizontalFov = atan (tan (verticalFov) * aspectratio)
	// verticalFov = atan (tan (horizontalFov) / aspectratio)
	if (aspect > (BASELINE_H / BASELINE_W))
	{
		// use the same calculation as GLQuake did (horizontal is constant, vertical varies)
		rd->fov_x = fovvar;
		rd->fov_y = SCR_CalcFovY (rd->fov_x, width, height);
	}
	else
	{
		// alternate calculation (vertical is constant, horizontal varies)
		// consistent with http://www.emsai.net/projects/widescreen/fovcalc/
		// note that the gun always uses this calculation irrespective of the aspect)
		rd->fov_y = SCR_CalcFovY (fovvar, BASELINE_W, BASELINE_H);
		rd->fov_x = SCR_CalcFovX (rd->fov_y, width, height);
	}
}


/*
=================
SCR_DrawCrosshair
=================
*/
void SCR_DrawCrosshair (void)
{
	if (!crosshair->value)
		return;

	if (crosshair->modified)
	{
		crosshair->modified = false;
		SCR_TouchPics ();
	}

	if (!crosshair_pic[0])
		return;

	re->DrawPic ((viddef.width - crosshair_width) >> 1, (viddef.height - crosshair_height) >> 1, crosshair_pic);
}


/*
==================
V_RenderView

==================
*/
void V_RenderView (void)
{
	extern int entitycmpfnc (const entity_t *, const entity_t *);

	if (cls.state != ca_active)
		return;

	if (!cl.refresh_prepped)
		return;			// still loading

	if (cl_timedemo->value)
	{
		if (!cl.timedemo_start)
			cl.timedemo_start = Sys_Milliseconds ();

		cl.timedemo_frames++;
	}

	// an invalid frame will just use the exact previous refdef
	// we can't use the old frame if the video mode has changed, though...
	if (cl.frame.valid && (cl.force_refdef || !cl_paused->value))
	{
		cl.force_refdef = false;

		V_ClearScene ();

		// build a refresh entity list and calc cl.sim*
		// this also calls CL_CalcViewValues which loads
		// v_forward, etc.
		CL_AddEntities ();

		if (cl_testparticles->value) V_TestParticles ();
		if (cl_testentities->value) V_TestEntities ();
		if (cl_testlights->value) V_TestLights ();

		if (cl_testblend->value)
		{
			cl.refdef.blend[0] = 1;
			cl.refdef.blend[1] = 0.5;
			cl.refdef.blend[2] = 0.25;
			cl.refdef.blend[3] = 0.5;
		}

		// never let it sit exactly on a node line, because a water plane can
		// dissapear when viewed with the eye exactly on it.
		// the server protocol only specifies to 1/8 pixel, so add 1/16 in each axis
		cl.refdef.vieworigin[0] += 1.0 / 16;
		cl.refdef.vieworigin[1] += 1.0 / 16;
		cl.refdef.vieworigin[2] += 1.0 / 16;

		// taken from proper vid.dimension, NOT the scaled console dimensions
		cl.refdef.x = 0;
		cl.refdef.y = 0;
		cl.refdef.width = viddef.width;
		cl.refdef.height = viddef.height;

		SCR_SetFOV (&cl.refdef, cl.refdef.fov_x, cl.refdef.width, cl.refdef.height);

		cl.refdef.time = cl.time * 0.001;
		cl.refdef.areabits = cl.frame.areabits;

		if (!cl_add_entities->value) r_numentities = 0;
		if (!cl_add_particles->value) r_numparticles = 0;
		if (!cl_add_lights->value) r_numdlights = 0;
		if (!cl_add_blend->value) Vector4Set (cl.refdef.blend, 0, 0, 0, 0);

		cl.refdef.num_entities = r_numentities;
		cl.refdef.entities = r_entities;
		cl.refdef.num_particles = r_numparticles;
		cl.refdef.particles = r_particles;
		cl.refdef.num_dlights = r_numdlights;
		cl.refdef.dlights = r_dlights;
		cl.refdef.lightstyles = r_lightstyles;

		cl.refdef.rdflags = cl.frame.playerstate.rdflags;

		// sort entities for better cache locality
		qsort (cl.refdef.entities, cl.refdef.num_entities, sizeof (cl.refdef.entities[0]), (sortfunc_t) entitycmpfnc);
	}

	re->RenderFrame (&cl.refdef);

	if (cl_stats->value)
		Com_Printf ("ent:%i  lt:%i  part:%i\n", r_numentities, r_numdlights, r_numparticles);

	re->Set2DMode ();

	SCR_DrawCrosshair ();
}


/*
=============
V_Viewpos_f
=============
*/
void V_Viewpos_f (void)
{
	Com_Printf ("(%i %i %i) : %i\n", (int) cl.refdef.vieworigin[0],
		(int) cl.refdef.vieworigin[1], (int) cl.refdef.vieworigin[2],
		(int) cl.refdef.viewangles[YAW]);
}

/*
=============
V_Init
=============
*/
void V_Init (void)
{
	Cmd_AddCommand ("viewpos", V_Viewpos_f);

	crosshair = Cvar_Get ("crosshair", "0", CVAR_ARCHIVE);

	cl_testblend = Cvar_Get ("cl_testblend", "0", 0);
	cl_testparticles = Cvar_Get ("cl_testparticles", "0", 0);
	cl_testentities = Cvar_Get ("cl_testentities", "0", 0);
	cl_testlights = Cvar_Get ("cl_testlights", "0", 0);
	cl_fov = Cvar_Get ("cl_fov", "90", CVAR_USERINFO | CVAR_ARCHIVE);
	cl_stats = Cvar_Get ("cl_stats", "0", 0);
}


