/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// in_win.c -- windows 95 mouse and joystick code

#include "quakedef.h"

#define DIRECTINPUT_VERSION 0x0800

#include <dinput.h>

#pragma comment (lib, "dinput8.lib")
#pragma comment (lib, "dxguid.lib")

cvar_t *in_mouse;

// this cvar only exists so that the menu won't crash
cvar_t *in_joystick;

// directinput objects
IDirectInput8 *di_Object = NULL;
IDirectInputDevice8 *di_Mouse = NULL;

qboolean in_appactive;
qboolean mlooking;
qboolean mouseactive;
qboolean mouseinitialized;

int			mouse_x[2];
int			mouse_y[2];

void IN_MLookDown (void) { mlooking = true; }
void IN_ClearStates (void);

void IN_MLookUp (void)
{
	mlooking = false;

	if (!freelook->value && lookspring->value)
		IN_CenterView ();
}


void IN_ShutdownDirectInput (void)
{
	if (di_Mouse)
	{
		di_Mouse->lpVtbl->Unacquire (di_Mouse);
		di_Mouse->lpVtbl->Release (di_Mouse);
		di_Mouse = NULL;
	}

	if (di_Object)
	{
		di_Object->lpVtbl->Release (di_Object);
		di_Object = NULL;
	}
}


/*
===========
IN_ActivateMouse

Called when the window gains focus or changes in some way
===========
*/
void IN_ActivateMouse (void)
{
	if (!mouseinitialized) return;

	if (!in_mouse->value)
	{
		IN_ShutdownDirectInput ();
		mouseactive = false;
		return;
	}

	if (mouseactive) return;

	// this can be called if we alt-tabbed away without previously deactivating the mouse so shut everything down if it's currently up
	IN_ShutdownDirectInput ();

	// everything must succeed
	if (SUCCEEDED (DirectInput8Create (GetModuleHandle (NULL), DIRECTINPUT_VERSION, &IID_IDirectInput8, (void **) &di_Object, NULL)))
	{
		if (SUCCEEDED (di_Object->lpVtbl->CreateDevice (di_Object, &GUID_SysMouse, &di_Mouse, NULL)))
		{
			if (SUCCEEDED (di_Mouse->lpVtbl->SetDataFormat (di_Mouse, &c_dfDIMouse2)))
			{
				if (SUCCEEDED (di_Mouse->lpVtbl->SetCooperativeLevel (di_Mouse, cl_hwnd, DISCL_FOREGROUND | DISCL_EXCLUSIVE)))
				{
					DIPROPDWORD	dipdw = {{sizeof (DIPROPDWORD), sizeof (DIPROPHEADER), 0, DIPH_DEVICE}, 16};

					if (SUCCEEDED (di_Mouse->lpVtbl->SetProperty (di_Mouse, DIPROP_BUFFERSIZE, &dipdw.diph)))
					{
						// the initial acquire is allowed to fail; it will try again the first time we read the mouse
						// this seems to play nicer with a certain touchpad manufacturer who's drivers rival AMD's for quality.
						di_Mouse->lpVtbl->Acquire (di_Mouse);
						mouseactive = true;
						return;
					}
				}
			}
		}
	}

	// something went wrong so release all di objects
	IN_ShutdownDirectInput ();
}


/*
===========
IN_DeactivateMouse

Called when the window loses focus
===========
*/
void IN_DeactivateMouse (void)
{
	if (!mouseinitialized) return;
	if (!mouseactive) return;

	// clear input state on a deactivate
	IN_ClearStates ();

	// shut down directinput
	IN_ShutdownDirectInput ();

	// now it's inactive
	mouseactive = false;
}


/*
===========
IN_StartupMouse
===========
*/
void IN_StartupMouse (void)
{
	cvar_t *cv = Cvar_Get ("in_initmouse", "1", CVAR_NOSET);

	if (!cv->value)
		return;

	mouseinitialized = true;
}


/*
===========
IN_MouseMove
===========
*/
void IN_MouseMove (usercmd_t *cmd)
{
	// add in the accumulated mouse movement
	// dinput has lower sensitivity and no ballistics so we need to boost it a little
	float mx = (float) (mouse_x[0] + mouse_x[1]) * sensitivity->value * 0.75f;
	float my = (float) (mouse_y[0] + mouse_y[1]) * sensitivity->value * 0.75f;

	// add mouse X/Y movement to cmd
	if ((in_strafe.state & 1) || (lookstrafe->value && mlooking))
		cmd->sidemove += m_side->value * mx;
	else cl.viewangles[YAW] -= m_yaw->value * mx;

	if ((mlooking || freelook->value) && !(in_strafe.state & 1))
		cl.viewangles[PITCH] += m_pitch->value * my;
	else cmd->forwardmove -= m_forward->value * my;

	// copy back to old state for filtering
	mouse_x[1] = mouse_x[0];
	mouse_y[1] = mouse_y[0];

	// clear the accumulated movement
	mouse_x[0] = 0;
	mouse_y[0] = 0;
}


void IN_SampleMouse (void)
{
	int mx = 0;
	int my = 0;
	DWORD dwDITime = 0;

	// the mouse is not currently in use
	if (!mouseactive) return;
	if (!di_Mouse) return;
	if (!di_Object) return;

	// even if we're inactive we always sample the mouse and respond to key events which keeps the
	// dinput buffers flushed so nothing backs up; we'll throw away the movement later if needed
	for (;;)
	{
		DWORD dwElements = 16;
		DIDEVICEOBJECTDATA di_Data[16];

		switch (di_Mouse->lpVtbl->GetDeviceData (di_Mouse, sizeof (DIDEVICEOBJECTDATA), di_Data, &dwElements, 0))
		{
		case DIERR_INPUTLOST:
		case DIERR_NOTACQUIRED:
			// device was lost; reacquire it
			di_Mouse->lpVtbl->Acquire (di_Mouse);
			return;

		case DI_OK:
			// we can read now
			break;

		default:
			// something else went wrong; attempt to recreate everything
			IN_DeactivateMouse ();
			Sleep (10);
			IN_ActivateMouse ();
			return;
		}

		if (dwElements < 1) break;

		// accumulate this move
		for (int i = 0; i < dwElements; i++)
		{
			DIDEVICEOBJECTDATA *od = &di_Data[i];

			// update message time
			if (od->dwTimeStamp > dwDITime) dwDITime = od->dwTimeStamp;

			// for x and y we must cast to remove the unsigned-ness for negative movements (because we're accumulating to floats)
			switch (od->dwOfs)
			{
			case DIMOFS_X:
				mx += (signed int) od->dwData;
				break;

			case DIMOFS_Y:
				my += (signed int) od->dwData;
				break;

			case DIMOFS_Z:
				if ((signed int) od->dwData > 0)
				{
					Key_Event (K_MWHEELUP, true, od->dwTimeStamp);
					Key_Event (K_MWHEELUP, false, od->dwTimeStamp);
				}
				else if ((signed int) od->dwData < 0)
				{
					Key_Event (K_MWHEELDOWN, true, od->dwTimeStamp);
					Key_Event (K_MWHEELDOWN, false, od->dwTimeStamp);
				}
				break;

				// respond to buttons
			case DIMOFS_BUTTON0: Key_Event (K_MOUSE1, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON1: Key_Event (K_MOUSE2, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON2: Key_Event (K_MOUSE3, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON3: Key_Event (K_MOUSE4, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON4: Key_Event (K_MOUSE5, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON5: Key_Event (K_MOUSE6, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON6: Key_Event (K_MOUSE7, od->dwData & 0x80, od->dwTimeStamp); break;
			case DIMOFS_BUTTON7: Key_Event (K_MOUSE8, od->dwData & 0x80, od->dwTimeStamp); break;
			}
		}

		// Com_Printf ("read %i elements\n", dwElements);
	}

	if (cls.state != ca_active)
	{
		// clear the accumulated movement
		mouse_x[0] = mouse_x[1] = 0;
		mouse_y[0] = mouse_y[1] = 0;
	}
	else
	{
		// add to the accumulated movement
		mouse_x[0] += mx;
		mouse_y[0] += my;
	}
}


cvar_t	*v_centermove;
cvar_t	*v_centerspeed;


/*
===========
IN_Init
===========
*/
void IN_Init (void)
{
	// mouse variables
	in_mouse = Cvar_Get ("in_mouse", "1", CVAR_ARCHIVE);

	// this cvar only exists so that the menu won't crash
	in_joystick = Cvar_Get ("in_joystick", "0", 0);

	// centering
	v_centermove = Cvar_Get ("v_centermove", "0.15", 0);
	v_centerspeed = Cvar_Get ("v_centerspeed", "500", 0);

	Cmd_AddCommand ("+mlook", IN_MLookDown);
	Cmd_AddCommand ("-mlook", IN_MLookUp);

	IN_StartupMouse ();
}

/*
===========
IN_Shutdown
===========
*/
void IN_Shutdown (void)
{
	IN_ClearStates ();
	IN_DeactivateMouse ();
}


/*
===========
IN_Activate

Called when the main window gains or loses focus.
The window may have been destroyed and recreated
between a deactivate and an activate.
===========
*/
void IN_Activate (qboolean active)
{
	in_appactive = active;
	mouseactive = !active;		// force a new window check or turn off
}


/*
==================
IN_Frame

Called every frame, even if not generating commands
==================
*/
void IN_Frame (void)
{
	if (!mouseinitialized)
		return;

	if (!in_mouse || !in_appactive)
	{
		IN_DeactivateMouse ();
		return;
	}

	if (cl.cinematictime > 0)
	{
		// hide the cursor in cinematics too
		IN_ActivateMouse ();
		return;
	}

	if (!cl.refresh_prepped || cls.key_dest == key_console || cls.key_dest == key_menu)
	{
		// temporarily deactivate if in a windowed mode
		if (Cvar_VariableValue ("vid_fullscreen") == 0)
		{
			IN_DeactivateMouse ();
			return;
		}
	}

	IN_ActivateMouse ();
}


/*
===================
IN_ClearStates
===================
*/
void IN_ClearStates (void)
{
	if (di_Mouse)
	{
		// flush the dinput buffers
		DWORD dwItems = INFINITE;

		// this is allowed to fail harmlessly
		di_Mouse->lpVtbl->GetDeviceData (
			di_Mouse,
			sizeof (DIDEVICEOBJECTDATA),
			NULL,
			&dwItems,
			0);
	}

	// send a key up event for any mouse keys that are down
	for (int i = 0; i < 8; i++)
		Key_Event (K_MOUSE1 + i, false, Sys_Milliseconds ());

	// clear accumulated mouse movement
	mouse_x[0] = mouse_x[1] = 0;
	mouse_y[0] = mouse_y[1] = 0;
}

