/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "gl_local.h"

typedef struct alphaobject_s
{
	entity_t *ent;
	msurface_t *surf;
	float alpha;
	float dist;
} alphaobject_t;

#define MAX_ALPHA_OBJECTS	65536

alphaobject_t r_alphaobjects[MAX_ALPHA_OBJECTS];
int r_numalphaobjects = 0;

// !!!!! if this is changed you must fixup "if (o->surf)" condition below !!!!!
msurface_t *r_alpha_surfaces[MAX_ALPHA_OBJECTS];
int r_num_alpha_surfaces = 0;

void R_ClearAlphaObjects (void)
{
	r_numalphaobjects = 0;
}


void R_AddAlphaObject (entity_t *e, msurface_t *surf, float alpha)
{
	if (r_numalphaobjects < MAX_ALPHA_OBJECTS)
	{
		float distvec[3];

		r_alphaobjects[r_numalphaobjects].ent = e;
		r_alphaobjects[r_numalphaobjects].surf = surf;
		r_alphaobjects[r_numalphaobjects].alpha = alpha;

		// evaluate distance to the entity for alpha sorting
		// this is a squared distance which will remain valid for comparison
		if (surf)
		{
			// only transform the surf if it needs to be transformed
			if (e->rotated)
			{
				// full matrix transform
				R_Transform (&e->localmatrix, distvec, surf->mids);
				Vector3Subtract (distvec, distvec, r_newrefdef.vieworigin);
			}
			else if (e->translated)
			{
				// just move the point
				Vector3Add (distvec, surf->mids, e->currorigin);
				Vector3Subtract (distvec, distvec, r_newrefdef.vieworigin);
			}
			else
			{
				// not transformed at all
				Vector3Subtract (distvec, surf->mids, r_newrefdef.vieworigin);
			}
		}
		else if (e->flags & RF_BEAM)
		{
			float beammids[3] = {
				(e->currorigin[0] + e->prevorigin[0]) * 0.5f,
				(e->currorigin[1] + e->prevorigin[1]) * 0.5f,
				(e->currorigin[2] + e->prevorigin[2]) * 0.5f
			};

			Vector3Subtract (distvec, beammids, r_newrefdef.vieworigin);
		}
		else Vector3Subtract (distvec, e->currorigin, r_newrefdef.vieworigin);

		// and now do the squared distance
		r_alphaobjects[r_numalphaobjects].dist = Vector3Dot (distvec, distvec);

		r_numalphaobjects++;
	}
}


int R_AlphaCompFunc (const alphaobject_t *a, const alphaobject_t *b)
{
	return (a->dist > b->dist) ? -1 : 1;
}


void R_DrawAlphaObjects (void)
{
	if (!r_numalphaobjects) return;

	// common render state (individual types may change this)
	R_SetRenderStates (RS_DEPTHTEST | RS_CULLCCW | RS_ALPHABLEND);

	// sort by dist
	qsort (r_alphaobjects, r_numalphaobjects, sizeof (alphaobject_t), (sortfunc_t) R_AlphaCompFunc);

	entity_t *r_lastent = NULL;
	float r_lastalpha = -666;

	// no alpha surfaces yet
	r_num_alpha_surfaces = 0;

	for (int i = 0; i < r_numalphaobjects; i++)
	{
		alphaobject_t *o = &r_alphaobjects[i];

		// figure entity change
		if (o->ent != r_lastent)
		{
			// flush alpha surfaces
			R_DrawAlphaSurfaces (r_alpha_surfaces, r_num_alpha_surfaces);
			r_num_alpha_surfaces = 0;

			// run the transform for the new entity
			R_SetEntityTransform (&o->ent->localmatrix, 0);
		}

		// figure alpha change
		if (o->alpha != r_lastalpha)
		{
			// flush alpha surfaces
			R_DrawAlphaSurfaces (r_alpha_surfaces, r_num_alpha_surfaces);
			r_num_alpha_surfaces = 0;

			// update alpha
			R_SetPixelShaderConstant4f (PSC_ALPHAVAL, 1, 1, 1, o->alpha);
		}

		// handle the current object
		if (o->surf)
		{
			// add to alpha surfaces
			// r_alpha_surfaces was sized at MAX_ALPHA_OBJECTS so this can be done safely without an overflow check
			r_alpha_surfaces[r_num_alpha_surfaces] = o->surf;
			r_num_alpha_surfaces++;
		}
		else
		{
			// flush alpha surfaces (is this ever needed here??? surely if the entity has changed
			// (it was a bmodel, now it's something else) they will already have been flushed???)
			R_DrawAlphaSurfaces (r_alpha_surfaces, r_num_alpha_surfaces);
			r_num_alpha_surfaces = 0;

			// draw this object
			R_DrawSingleEntityOnList (o->ent);
		}

		// store back
		r_lastalpha = o->alpha;
		r_lastent = o->ent;
	}

	// flush alpha surfaces
	R_DrawAlphaSurfaces (r_alpha_surfaces, r_num_alpha_surfaces);

	// remove the alpha
	R_SetPixelShaderConstant4f (PSC_ALPHAVAL, 1, 1, 1, 1);

	// for the next frame
	r_num_alpha_surfaces = 0;
	r_numalphaobjects = 0;
}


