/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c

#include "gl_local.h"
#include "resource.h"

typedef struct drawpolyvert_s
{
	float xy[2];
	uint32_t c;
	float st[2];
} drawpolyvert_t;

typedef struct postpolyvert_s
{
	float xy[2];
	float st0[2];
	float st1[2];
} postpolyvert_t;

__inline static void Draw_AddVertex (drawpolyvert_t *v, float x, float y, unsigned c, float s, float t)
{
	v->xy[0] = x;
	v->xy[1] = y;
	v->c = c;
	v->st[0] = s;
	v->st[1] = t;
}

__inline static void Post_AddVertex (postpolyvert_t *v, float x, float y, float s0, float t0, float s1, float t1)
{
	v->xy[0] = x;
	v->xy[1] = y;
	v->st0[0] = s0;
	v->st0[1] = t0;
	v->st1[0] = s1;
	v->st1[1] = t1;
}


// all images except character rects will use these
static texrect_t r_standardtexcoords = {0, 1, 0, 1};

IDirect3DVertexDeclaration9 *d3d_DrawDecl = NULL;
IDirect3DVertexDeclaration9 *d3d_PostDecl = NULL;

IDirect3DVertexShader9 *d3d_DrawTexturedVS = NULL;
IDirect3DPixelShader9 *d3d_DrawTexturedPS = NULL;

IDirect3DVertexShader9 *d3d_DrawColouredVS = NULL;
IDirect3DPixelShader9 *d3d_DrawColouredPS = NULL;

IDirect3DVertexShader9 *d3d_DrawPolyBlendVS = NULL;
IDirect3DPixelShader9 *d3d_DrawPolyBlendPS = NULL;

IDirect3DVertexShader9 *d3d_PostProcessVS = NULL;
IDirect3DVertexBuffer9 *d3d_PostProcessVB = NULL;

image_t d3d_RawTexture;
image_t d3d_PostProcessTexture;


void Draw_OnCreate (void)
{
	D3DVERTEXELEMENT9 drawlayout[] = {
		VDECL (0, offsetof (drawpolyvert_t, xy), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_POSITION, 0),
		VDECL (0, offsetof (drawpolyvert_t, c), D3DDECLTYPE_D3DCOLOR, D3DDECLUSAGE_COLOR, 0),
		VDECL (0, offsetof (drawpolyvert_t, st), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 0),
		D3DDECL_END ()
	};

	D3DVERTEXELEMENT9 postlayout[] = {
		VDECL (0, offsetof (postpolyvert_t, xy), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_POSITION, 0),
		VDECL (0, offsetof (postpolyvert_t, st0), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 0),
		VDECL (0, offsetof (postpolyvert_t, st1), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 1),
		D3DDECL_END ()
	};

	d3d_DrawDecl = R_CreateVertexDeclaration (drawlayout);
	d3d_PostDecl = R_CreateVertexDeclaration (postlayout);

	R_CreateShaders (DrawTexturedVS, DrawTexturedPS, IDR_DRAWSHADER);
	R_CreateShaders (DrawColouredVS, DrawColouredPS, IDR_DRAWSHADER);
	R_CreateShaders (DrawPolyBlendVS, DrawPolyBlendPS, IDR_DRAWSHADER);

	d3d_PostProcessVS = R_CreateVertexShader (IDR_DRAWSHADER, "PostProcessVS");
}


void Draw_OnDestroy (void)
{
	R_ReleaseTexture (&d3d_RawTexture);
	SAFE_RELEASE (d3d_PostProcessVB);
}


void Draw_OnReset (void)
{
	// create a texture to use as a post-process target
	// because we want this sized to the video width/height we create it manually instead of sending it through R_CreateTexture
	// if it fails because of e.g lack of NP2 support we just don't run post-process passes
	// it must be created as a rendertarget or stretchrect will fail
	IDirect3DSurface9 *bbuff = NULL;
	D3DSURFACE_DESC desc;

	// take it's dimensions from the backbuffer so that it's properly sized
	if (SUCCEEDED (d3d_Device->lpVtbl->GetBackBuffer (d3d_Device, 0, 0, D3DBACKBUFFER_TYPE_MONO, &bbuff)))
	{
		if (SUCCEEDED (bbuff->lpVtbl->GetDesc (bbuff, &desc)))
		{
			d3d_Device->lpVtbl->CreateTexture (
				d3d_Device,
				desc.Width,
				desc.Height,
				1,
				D3DUSAGE_RENDERTARGET,
				D3DFMT_X8R8G8B8,
				D3DPOOL_DEFAULT,
				&d3d_PostProcessTexture.texture,
				NULL);

			// tell the rest of the engine code about the new dimensions
			ri->Vid_NewWindow (desc.Width, desc.Height);

			// create the VB while we've got the desc so that we can size the texcoords correctly
			if (SUCCEEDED (d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, sizeof (postpolyvert_t) * 4, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &d3d_PostProcessVB, NULL)))
			{
				postpolyvert_t *verts = NULL;

				if (SUCCEEDED (d3d_PostProcessVB->lpVtbl->Lock (d3d_PostProcessVB, 0, 0, (void **) &verts, 0)))
				{
					// do it this way so it fills the viewport properly (fixme - put in a buffer)
					Post_AddVertex (&verts[0], -1, -1, 0, 1, 0, 1);
					Post_AddVertex (&verts[1],  1, -1, 1, 1, (float) desc.Width / (float) desc.Height, 1);
					Post_AddVertex (&verts[2], -1,  1, 0, 0, 0, 0);
					Post_AddVertex (&verts[3],  1,  1, 1, 0, (float) desc.Width / (float) desc.Height, 0);

					d3d_PostProcessVB->lpVtbl->Unlock (d3d_PostProcessVB);
				}
				else
				{
					// if the lock failed then it's not valid to do post-processing
					R_ReleaseTexture (&d3d_PostProcessTexture);
					SAFE_RELEASE (d3d_PostProcessVB);
				}
			}
		}

		bbuff->lpVtbl->Release (bbuff);
	}

	if (!d3d_PostProcessTexture.texture || !d3d_PostProcessVB)
	{
		// if either failed then it's not valid to do post-processing
		R_ReleaseTexture (&d3d_PostProcessTexture);
		SAFE_RELEASE (d3d_PostProcessVB);
	}
}


void Draw_OnLost (void)
{
	// we must release this too in order to invalidate the cached states
	R_ReleaseTexture (&d3d_RawTexture);

	// and this is a render target in the default pool
	R_ReleaseTexture (&d3d_PostProcessTexture);

	// this must also be resized/changed/etc
	SAFE_RELEASE (d3d_PostProcessVB);
}


void Draw_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Draw_OnCreate, Draw_OnLost, Draw_OnReset, Draw_OnDestroy);
		registered = true;
	}
}


void Draw_SetScale (float scale)
{
}


/*
=============
Draw_FindPic
=============
*/
image_t	*Draw_FindPic (char *name)
{
	image_t *gl;
	char	fullname[MAX_QPATH];

	if (name[0] != '/' && name[0] != '\\')
	{
		Com_sprintf (fullname, sizeof (fullname), "pics/%s.pcx", name);
		gl = R_FindImage (fullname, it_pic);
	}
	else gl = R_FindImage (name + 1, it_pic);

	return gl;
}


image_t	*draw_chars;
texrect_t charrects[256];

// allow the rendering code to pre-cache the little sb number pics so that we don't have to do texture cache lookups every time we want to use one
#define STAT_MINUS		10	// num frame for '-' stats digit

static char *sb_nums[2][11] =
{
	{
		"num_0", "num_1", "num_2", "num_3", "num_4", "num_5",
		"num_6", "num_7", "num_8", "num_9", "num_minus"
	},
	{
		"anum_0", "anum_1", "anum_2", "anum_3", "anum_4", "anum_5",
		"anum_6", "anum_7", "anum_8", "anum_9", "anum_minus"
	}
};

static image_t *sb_num_images[2][11];

#define	ICON_WIDTH	24
#define	ICON_HEIGHT	24
#define	CHAR_WIDTH	16
#define	ICON_SPACE	8

void Draw_SetupCharRects (float scale, float size)
{
	// set up pre-calculated character rectangles
	for (int num = 0; num < 256; num++)
	{
		charrects[num].sl = (float) (num & 15) * scale;
		charrects[num].sh = charrects[num].sl + size;
		charrects[num].tl = (float) (num >> 4) * scale;
		charrects[num].th = charrects[num].tl + size;
	}
}


void Draw_CopyBlock (byte *srcblock, byte *dstblock)
{
	for (int i = 0; i < 8; i++, srcblock += 128, dstblock += 256)
	{
		dstblock[0] = srcblock[0];
		dstblock[1] = srcblock[1];
		dstblock[2] = srcblock[2];
		dstblock[3] = srcblock[3];
		dstblock[4] = srcblock[4];
		dstblock[5] = srcblock[5];
		dstblock[6] = srcblock[6];
		dstblock[7] = srcblock[7];
	}
}


byte *Draw_GetBlock (byte *src, int row, int col, int stride, int size)
{
	byte *rowpointer = src + row * stride * size;
	return rowpointer + col * size;
}


void Draw_JoinWidgets (byte *src, byte *dst, int row, int col)
{
	Draw_CopyBlock (Draw_GetBlock (src, row, col, 128, 8), Draw_GetBlock (dst, row, col, 256, 16) + 8);
	Draw_CopyBlock (Draw_GetBlock (src, row, col, 128, 8), Draw_GetBlock (dst, row, col, 256, 16) - 8);
}


void Draw_PadCharset (byte *src, byte *dst)
{
	// the initial pad just expands out the charset to the new larger size
	for (int y = 0; y < 16; y++)
		for (int x = 0; x < 16; x++)
			Draw_CopyBlock (Draw_GetBlock (src, y, x, 128, 8), Draw_GetBlock (dst, y, x, 256, 16));

	// then we join together the various widgets wherever possible
	Draw_JoinWidgets (src, dst, 0, 2);	// textbox top
	Draw_JoinWidgets (src, dst, 0, 5);	// textbox middle
	Draw_JoinWidgets (src, dst, 0, 8);	// textbox bottom

	Draw_JoinWidgets (src, dst, 1, 3);	// field top
	Draw_JoinWidgets (src, dst, 1, 6);	// field middle
	Draw_JoinWidgets (src, dst, 1, 9);	// field bottom
	Draw_JoinWidgets (src, dst, 1, 14);	// console separator 1

	Draw_JoinWidgets (src, dst, 8, 1);	// scroll bar
	Draw_JoinWidgets (src, dst, 8, 3);	// selected item
	Draw_JoinWidgets (src, dst, 9, 14);	// console separator 2

	// finally do a bunch of memcpys from dst to dst to pad out the textbox and field chars
	// the first of these abuses texture wrap mode by copying to the bottom
	memcpy (&dst[256 * 255 + 64], &dst[256 * 0 + 64], 88);
	memcpy (&dst[256 * 8 + 16], &dst[256 * 7 + 16], 88);

	memcpy (&dst[256 * 15 + 80], &dst[256 * 16 + 80], 88);
	memcpy (&dst[256 * 24 + 32], &dst[256 * 23 + 32], 88);
}


/*
===============
Draw_InitLocal
===============
*/
void Draw_InitLocal (void)
{
	// precache all the little sbar numbers so that we don't need to search the image cache for them at runtime
	for (int c = 0; c < 2; c++)
		for (int n = 0; n < 11; n++)
			sb_num_images[c][n] = Draw_FindPic (sb_nums[c][n]);

	// try to load console characters by hand because we need to pad out the charset for supporting high values of scr_conscale
	byte *pal, *pic;
	int mark = ri->Scratch_LowMark ();
	int width, height;

	R_LoadPCX ("pics/conchars.pcx", &pic, &pal, &width, &height, R_Malloc);

	// sanity check that we've got a valid charset pic
	if (pic && pal && width == 128 && height == 128)
	{
		// alloc a new buffer for the chars to go into
		byte *newchars = (byte *) ri->Scratch_AllocSize (width * height * 4);

		// clear to alpha
		memset (newchars, 255, width * height * 4);

		// pad the old chars out to the new chars
		Draw_PadCharset (pic, newchars);

		// and load 'em
		draw_chars = R_LoadPic ("pics/conchars.pcx", newchars, 256, 256, it_pic, R_Palette24To32 (pal));
		Draw_SetupCharRects (0.0625f, 0.03125f);
	}
	else
	{
		// just load them the old way; if r_drawsize looks crap that's not my problem
		draw_chars = R_FindImage ("pics/conchars.pcx", it_pic);
		Draw_SetupCharRects (0.0625f, 0.0625f);
	}

	ri->Scratch_FreeToLowMark (mark);
}


// !!!!! don't even *THINK* of doing clamp mode on the draw stuff
static void Draw_TexturedQuad (int xl, int xh, int yl, int yh, int c, texrect_t *r)
{
	// we could batch these but we won't because there are only AT MOST ~8 to ~12 in any given frame
	drawpolyvert_t verts[4];

	Draw_AddVertex (&verts[0], xl, yl, c, r->sl, r->tl);
	Draw_AddVertex (&verts[1], xh, yl, c, r->sh, r->tl);
	Draw_AddVertex (&verts[2], xl, yh, c, r->sl, r->th);
	Draw_AddVertex (&verts[3], xh, yh, c, r->sh, r->th);

	R_SetDPUPMode ();
	R_SetVertexShader (d3d_DrawTexturedVS);
	R_SetPixelShader (d3d_DrawTexturedPS);

	d3d_Device->lpVtbl->DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLESTRIP, 2, verts, sizeof (drawpolyvert_t));
	c_draw_call++;
}


void Draw_Field (int x, int y, int color, int width, int value)
{
	char	num[16], *ptr;
	int		l;
	int		frame;

	if (width < 1)
		return;

	// draw number string
	if (width > 5)
		width = 5;

	Com_sprintf (num, sizeof (num), "%i", value);
	l = strlen (num);

	if (l > width)
		l = width;

	x += 2 + CHAR_WIDTH * (width - l);

	ptr = num;
	color &= 1;

	while (*ptr && l)
	{
		if (*ptr == '-')
			frame = STAT_MINUS;
		else frame = *ptr - '0';

		// retrieve the image from pre-cache so we don't need to go finding each one
		image_t *gl = sb_num_images[color][frame];

		// don't bother batching; a typical sbar layout only has at most 9-12 of these
		R_SetTexture (TS_DIFFUSE, gl);
		Draw_TexturedQuad (x, x + gl->width, y, y + gl->height, 0xffffffff, &r_standardtexcoords);

		x += CHAR_WIDTH;
		ptr++;
		l--;
	}
}


/*
================
Draw_Char

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Char (int x, int y, int num)
{
	num &= 255;

	if ((num & 127) == 32) return;	// space
	if (y <= -8) return;			// totally off screen

	// no point in batching as it's just used for blinking cursors/etc nowadays
	R_SetTexture (TS_DIFFUSE, draw_chars);
	Draw_TexturedQuad (x, x + 8, y, y + 8, 0xffffffff, &charrects[num]);
}


void Draw_String (int x, int y, const char *str, int len, int xor)
{
	drawpolyvert_t *verts = (drawpolyvert_t *) ri->Scratch_Alloc ();
	int num, numverts = 0;

	// totally off screen
	if (y <= -8) return;

	// get the length which is either specified (to chop a non-null-termed string) or the length of a null-termed string
	if (!str) return;
	if (len < 1) len = strlen (str);

	// remove leading spaces - can't use *str++ as it will always chop the first char even if it's non-space
	while (((num = *str ^ xor) & 127) == 32)
	{
		str++;
		x += 8;
		len--;
	}

	// don't draw zero-length strings
	if (len < 1) return;

	for (int i = 0; i < len; i++, x += 8)
	{
		if (!str[i]) break;
		if (((num = (str[i] & 255) ^ xor) & 127) == 32) continue;	// space

		// draw as 2 triangles so we don't need to maintain an index buffer
		Draw_AddVertex (&verts[0], x, y, 0xffffffff, charrects[num].sl, charrects[num].tl);
		Draw_AddVertex (&verts[1], x + 8, y, 0xffffffff, charrects[num].sh, charrects[num].tl);
		Draw_AddVertex (&verts[2], x + 8, y + 8, 0xffffffff, charrects[num].sh, charrects[num].th);

		// and add the second triangle
		Draw_AddVertex (&verts[3], x, y, 0xffffffff, charrects[num].sl, charrects[num].tl);
		Draw_AddVertex (&verts[4], x + 8, y + 8, 0xffffffff, charrects[num].sh, charrects[num].th);
		Draw_AddVertex (&verts[5], x, y + 8, 0xffffffff, charrects[num].sl, charrects[num].th);

		// advance here, not in the for loop, cos we may be continuing above if we have a space
		verts += 6;
		numverts += 6;
	}

	// nothing was drawn
	if (!numverts) return;

	// now setup to draw
	R_SetDPUPMode ();

	// states/etc
	R_SetVertexShader (d3d_DrawTexturedVS);
	R_SetPixelShader (d3d_DrawTexturedPS);
	R_SetTexture (TS_DIFFUSE, draw_chars);

	// each triangle is 3 verts
	d3d_Device->lpVtbl->DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLELIST, numverts / 3, ri->Scratch_Alloc (), sizeof (drawpolyvert_t));
	c_draw_call++;
}


/*
=============
Draw_GetPicSize

this was documented as returning 0, 0 for height and width if the pic wasn't found.  unfortunately it always
returned -1, -1 instead.  even more unfortunately there were places in the code that depended on it returning
0, 0, and they were hacked around and crutched-up in the original.  it's fixed now and all that mess is gone.
=============
*/
qboolean Draw_GetPicSize (int *w, int *h, char *pic)
{
	image_t *gl = Draw_FindPic (pic);

	if (!gl)
	{
		*w = *h = -1;
		return false;
	}

	*w = gl->width;
	*h = gl->height;

	return true;
}

/*
=============
Draw_StretchPic
=============
*/
void Draw_StretchPic (int x, int y, int w, int h, char *pic)
{
	image_t *gl = Draw_FindPic (pic);

	if (!gl)
	{
		ri->Con_Printf (PRINT_ALL, "Can't find pic: %s\n", pic);
		return;
	}

	R_SetTexture (TS_DIFFUSE, gl);
	Draw_TexturedQuad (x, x + w, y, y + h, 0xffffffff, &r_standardtexcoords);
}


void Draw_ConsoleBackground (int x, int y, int w, int h, char *pic, float alpha)
{
	// this restores the original Quake 1 behaviour of a translucent console background in selected views
	image_t *gl = Draw_FindPic (pic);
	unsigned rgba = ((int) (alpha * 255) << 24) + (255 << 0) + (255 << 8) + (255 << 16);

	if (!gl)
	{
		ri->Con_Printf (PRINT_ALL, "Can't find pic: %s\n", pic);
		return;
	}

	R_SetTexture (TS_DIFFUSE, gl);
	Draw_TexturedQuad (x, x + w, y, y + h, rgba, &r_standardtexcoords);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, char *pic)
{
	image_t *gl = Draw_FindPic (pic);

	if (!gl)
	{
		ri->Con_Printf (PRINT_ALL, "Can't find pic: %s\n", pic);
		return;
	}

	R_SetTexture (TS_DIFFUSE, gl);
	Draw_TexturedQuad (x, x + gl->width, y, y + gl->height, 0xffffffff, &r_standardtexcoords);
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	// instead of drawing a quad we'll clear a screen rect (which may be cheaper)
	D3DRECT r = {
		(x * vid.width) / vid.width,
		(y * vid.height) / vid.height,
		((x + w) * vid.width) / vid.width,
		((y + h) * vid.height) / vid.height
	};

	d3d_Device->lpVtbl->Clear (d3d_Device, 1, &r, D3DCLEAR_TARGET, d_8to24table[c & 255], 1.0f, 0);
}

void Draw_Clear (unsigned colour)
{
	d3d_Device->lpVtbl->Clear (d3d_Device, 0, NULL, D3DCLEAR_TARGET, d_8to24table[colour & 255], 1.0f, 0);
}


//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	// this is the only coloured quad left in the engine
	unsigned c = 0xcc000000;
	drawpolyvert_t verts[4] = {{{-1, -1}, c}, {{1, -1}, c}, {{-1, 1}, c}, {{1, 1}, c}};

	R_SetDPUPMode ();
	R_SetVertexShader (d3d_DrawColouredVS);
	R_SetPixelShader (d3d_DrawColouredPS);

	d3d_Device->lpVtbl->DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLESTRIP, 2, verts, sizeof (drawpolyvert_t));
	c_draw_call++;
}


/*
=============
Draw_StretchRaw
=============
*/
static unsigned r_rawpalette[256];

void R_SetCinematicPalette (const uint8_t *palette)
{
	if (palette)
		memcpy (r_rawpalette, R_Palette24To32 (palette), sizeof (r_rawpalette));
	else memcpy (r_rawpalette, d_8to24table, sizeof (r_rawpalette));

	// refresh the texture if the palette also changes
	Draw_StretchRaw (0, 0, NULL, -1);
}


void Draw_StretchRaw (int cols, int rows, byte *data, int frame)
{
	static int r_rawframe = -1;

	if (cols < 1 || rows < 1 || !data || frame < 0)
	{
		// this is a "reset" mode that just forces a recache of the texture next time it's drawn
		r_rawframe = -1;
		return;
	}

	if (d3d_RawTexture.width != cols || d3d_RawTexture.height != rows)
	{
		R_ReleaseTexture (&d3d_RawTexture);
		d3d_RawTexture.width = cols;
		d3d_RawTexture.height = rows;
	}

	if (!d3d_RawTexture.texture)
	{
		d3d_RawTexture.texture = R_CreateTexture (cols, rows, false, false);
		r_rawframe = -1;
	}

	if (r_rawframe != frame)
	{
		R_FillTexture (d3d_RawTexture.texture, data, cols, rows, r_rawpalette, false);
		r_rawframe = frame;
	}

	// no point in batching this as it's always only ever a single quad
	R_SetTexture (TS_DIFFUSE, &d3d_RawTexture);

	// stretch up the pic to fill the viewport while still maintaining aspect
	int w = vid.width;
	int h = (w * rows) / cols;

	if (h > vid.height)
	{
		h = vid.height;
		w = (h * cols) / rows;
	}

	// take it down if required
	while (w > vid.width) w >>= 1;
	while (h > vid.height) h >>= 1;

	// and center it properly
	int x = (vid.width - w) >> 1;
	int y = (vid.height - h) >> 1;

	texrect_t texcoords = {0, 1, 0, 1};

	// switch the sampler to linear/clamp always
	R_SetSamplerState (TS_DIFFUSE, D3DTEXF_LINEAR, D3DTEXF_NONE, D3DTADDRESS_CLAMP, 1);

	// and draw it
	Draw_TexturedQuad (x, x + w, y, y + h, 0xffffffff, &texcoords);

	// put the sampler back
	R_TextureMode (gl_texturemode->string, (int) gl_textureanisotropy->value);
}


void R_SetGL2D (void)
{
	// let's set up our viewport for drawing...
	// this forces all 2D drawing to the very front of the Z-order
	D3DVIEWPORT9 vp = {0, 0, vid.width, vid.height, 0, 0};
	R_SetViewport (&vp);

	// set up the draw matrix
	glmatrix m;
	R_IdentityMatrix (&m);
	R_OrthoMatrix (&m, 0, vid.width, vid.height, 0, -1, 1);

	// this matrix always goes into slot 32
	R_SetVertexShaderConstant4fv (VSC_ORTHOMATRIX, m.m16, 4);

	// !!!!! don't even *THINK* of doing clamp mode on the draw stuff
	R_SetRenderStates (RS_ALPHATEST | RS_ALPHABLEND);
	R_SetVertexDeclaration (d3d_DrawDecl);
}


void Draw_PolyBlend (float *color)
{
	// do it this way so it fills the viewport properly
	drawpolyvert_t verts[4] = {{{-1, -1}}, {{1, -1}}, {{-1, 1}}, {{1, 1}}};

	R_SetDPUPMode ();
	R_SetRenderStates (RS_ALPHABLEND);

	R_SetVertexDeclaration (d3d_DrawDecl);
	R_SetVertexShader (d3d_DrawPolyBlendVS);
	R_SetPixelShader (d3d_DrawPolyBlendPS);

	R_SetPixelShaderConstant4fv (PSC_POLYBLEND, color, 1);
	d3d_Device->lpVtbl->DrawPrimitiveUP (d3d_Device, D3DPT_TRIANGLESTRIP, 2, verts, sizeof (drawpolyvert_t));
	c_draw_call++;
}


void Draw_RunPostProcessStep (funcptr_t PostProcessCallBack)
{
	// generic function for running post-processing stages; copies the screen, sets up verts/etc, runs a callback to set any other
	// required state, then issues the call.  the callback should at the minimum specify a pixel shader and it may set other state
	// try to copy over the backbuffer to the gamma surface
	IDirect3DSurface9 *gsurf = NULL;
	IDirect3DSurface9 *bbuff = NULL;

	// if the texture or VB didn't create, or if no callback was specified then we just don't do it
	if (!d3d_PostProcessTexture.texture) return;
	if (!PostProcessCallBack) return;
	if (!d3d_PostProcessVB) return;

	if (SUCCEEDED (d3d_PostProcessTexture.texture->lpVtbl->GetSurfaceLevel (d3d_PostProcessTexture.texture, 0, &gsurf)))
	{
		if (SUCCEEDED (d3d_Device->lpVtbl->GetRenderTarget (d3d_Device, 0, &bbuff)))
		{
			// copy it over
			if (SUCCEEDED (d3d_Device->lpVtbl->StretchRect (d3d_Device, bbuff, NULL, gsurf, NULL, D3DTEXF_NONE)))
			{
				// only draw it if the copy succeeded
				R_SetRenderStates (0);
				R_SetVertexShader (d3d_PostProcessVS);
				R_SetTexture (TS_SCREEN, &d3d_PostProcessTexture);
				R_SetVertexDeclaration (d3d_PostDecl);
				R_SetVertexBuffer (0, d3d_PostProcessVB, 0, sizeof (postpolyvert_t));

				// issue the callback to set up any other required state
				PostProcessCallBack ();

				// and issue the draw call
				d3d_Device->lpVtbl->DrawPrimitive (d3d_Device, D3DPT_TRIANGLESTRIP, 0, 2);
				c_draw_call++;
			}

			bbuff->lpVtbl->Release (bbuff);
		}

		gsurf->lpVtbl->Release (gsurf);
	}
}

