/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_mesh.c: triangle model functions

#include "gl_local.h"
#include "resource.h"

typedef struct vertexnormals_s
{
	int numnormals;
	float normal[3];
} vertexnormals_t;

typedef struct aliaspolyvert_s
{
	byte position[4];
	float normal[3];
} aliaspolyvert_t;

typedef struct aliastexcoord_s
{
	float st[2];
} aliastexcoord_t;


IDirect3DVertexDeclaration9 *d3d_MeshDecl = NULL;

IDirect3DVertexShader9 *d3d_MeshTexturedVS = NULL;
IDirect3DPixelShader9 *d3d_MeshTexturedPS = NULL;

IDirect3DVertexShader9 *d3d_MeshShellVS = NULL;
IDirect3DPixelShader9 *d3d_MeshShellPS = NULL;

IDirect3DVertexShader9 *d3d_MeshDynamicVS = NULL;
IDirect3DPixelShader9 *d3d_MeshDynamicPS = NULL;

typedef struct meshbuffers_s
{
	IDirect3DVertexBuffer9 *positions;
	IDirect3DVertexBuffer9 *texcoords;
	IDirect3DIndexBuffer9 *indexes;
	int registration_sequence;
	char name[MAX_PATH];
	int numverts;
	int numindexes;
	int numtris;
} meshbuffers_t;

#define MAX_MESH_BUFFERS	1024

meshbuffers_t d3d_MeshBuffers[MAX_MESH_BUFFERS];


void Mesh_OnCreate (void)
{
	D3DVERTEXELEMENT9 layout[] = {
		// because some of the boss models are so huge we compress the position data
		VDECL (0, offsetof (aliaspolyvert_t, position), D3DDECLTYPE_UBYTE4, D3DDECLUSAGE_POSITION, 0),
		VDECL (0, offsetof (aliaspolyvert_t, normal), D3DDECLTYPE_FLOAT3, D3DDECLUSAGE_NORMAL, 0),
		VDECL (1, offsetof (aliaspolyvert_t, position), D3DDECLTYPE_UBYTE4, D3DDECLUSAGE_POSITION, 1),
		VDECL (1, offsetof (aliaspolyvert_t, normal), D3DDECLTYPE_FLOAT3, D3DDECLUSAGE_NORMAL, 1),
		VDECL (2, offsetof (aliastexcoord_t, st), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 0),
		D3DDECL_END ()
	};

	d3d_MeshDecl = R_CreateVertexDeclaration (layout);

	R_CreateShaders (MeshTexturedVS, MeshTexturedPS, IDR_MESHSHADER);
	R_CreateShaders (MeshShellVS, MeshShellPS, IDR_MESHSHADER);
	R_CreateShaders (MeshDynamicVS, MeshDynamicPS, IDR_MESHSHADER);
}


void Mesh_OnReset (void)
{
}


void Mesh_OnLost (void)
{
}


void Mesh_OnDestroy (void)
{
	for (int i = 0; i < MAX_MESH_BUFFERS; i++)
	{
		SAFE_RELEASE (d3d_MeshBuffers[i].positions);
		SAFE_RELEASE (d3d_MeshBuffers[i].texcoords);
		SAFE_RELEASE (d3d_MeshBuffers[i].indexes);

		memset (&d3d_MeshBuffers[i], 0, sizeof (d3d_MeshBuffers[i]));
	}
}

void Mesh_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Mesh_OnCreate, Mesh_OnLost, Mesh_OnReset, Mesh_OnDestroy);
		registered = true;
	}
}


void R_BuildFrameNormals (dmdl_t *hdr, daliasframe_t *frame, vertexnormals_t *vnorms)
{
	dtriangle_t *pintri = (dtriangle_t *) ((byte *) hdr + hdr->ofs_tris);

	for (int i = 0; i < hdr->num_xyz; i++)
	{
		// no normals initially
		Vector3Clear (vnorms[i].normal);
		vnorms[i].numnormals = 0;
	}

	for (int i = 0; i < hdr->num_tris; i++)
	{
		float triverts[3][3];
		float vtemp1[3], vtemp2[3], normal[3];
		short *vertindexes = pintri[i].index_xyz;

		// undo the vertex rotation from modelgen.c here too
		for (int j = 0; j < 3; j++)
		{
			triverts[j][0] = (float) frame->verts[vertindexes[j]].v[1] * frame->scale[1] + frame->translate[1];
			triverts[j][1] = -((float) frame->verts[vertindexes[j]].v[0] * frame->scale[0] + frame->translate[0]);
			triverts[j][2] = (float) frame->verts[vertindexes[j]].v[2] * frame->scale[2] + frame->translate[2];
		}

		// calc the per-triangle normal
		Vector3Subtract (vtemp1, triverts[0], triverts[1]);
		Vector3Subtract (vtemp2, triverts[2], triverts[1]);
		Vector3Cross (normal, vtemp1, vtemp2);
		Vector3Normalize (normal);

		// and accumulate it into the calculated normals array
		for (int j = 0; j < 3; j++)
		{
			// rotate the normal so the model faces down the positive x axis
			vnorms[vertindexes[j]].normal[0] -= normal[1];
			vnorms[vertindexes[j]].normal[1] += normal[0];
			vnorms[vertindexes[j]].normal[2] += normal[2];

			vnorms[vertindexes[j]].numnormals++;
		}
	}

	// copy out normals
	for (int i = 0; i < hdr->num_xyz; i++)
	{
		// numnormals was checked for > 0 in modelgen.c so we shouldn't need to do it again 
		// here but we do anyway just in case a rogue modder has used a bad modelling tool
		if (vnorms[i].numnormals > 0)
		{
			Vector3Scalef (vnorms[i].normal, vnorms[i].normal, (float) vnorms[i].numnormals);
			Vector3Normalize (vnorms[i].normal);
		}
		else Vector3Set (vnorms[i].normal, 0.0f, 0.0f, 1.0f);
	}
}


void R_CreateFramesBuffer (dmdl_t *hdr, meshbuffers_t *set)
{
	if (SUCCEEDED (d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, sizeof (aliaspolyvert_t) * set->numverts * hdr->num_frames, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &set->positions, NULL)))
	{
		aliaspolyvert_t *verts = NULL;

		if (SUCCEEDED (set->positions->lpVtbl->Lock (set->positions, 0, 0, (void **) &verts, 0)))
		{
			for (int f = 0; f < hdr->num_frames; f++)
			{
				vertexnormals_t *vnorms = (vertexnormals_t *) ri->Scratch_Alloc ();
				int *order = (int *) ((byte *) hdr + hdr->ofs_glcmds);
				daliasframe_t *frame = (daliasframe_t *) ((byte *) hdr + hdr->ofs_frames + f * hdr->framesize);

				// recalculate the normals for this frame properly
				R_BuildFrameNormals (hdr, frame, vnorms);

				// and now build the data for this frame
				for (;;)
				{
					// get the vertex count and primitive type
					int count = *order++;

					if (!count) break;
					if (count < 0) count = -count;

					for (int i = 0; i < count; i++, verts++, order += 3)
					{
						// copy over the position
						verts->position[0] = frame->verts[order[2]].v[0];
						verts->position[1] = frame->verts[order[2]].v[1];
						verts->position[2] = frame->verts[order[2]].v[2];
						verts->position[3] = 1;

						// copy over the normal
						verts->normal[0] = vnorms[order[2]].normal[0];
						verts->normal[1] = vnorms[order[2]].normal[1];
						verts->normal[2] = vnorms[order[2]].normal[2];
					}
				}
			}

			set->positions->lpVtbl->Unlock (set->positions);
		}
	}
}


void R_CreateTexCoordsBuffer (dmdl_t *hdr, meshbuffers_t *set)
{
	if (SUCCEEDED (d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, sizeof (aliastexcoord_t) * set->numverts, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &set->texcoords, NULL)))
	{
		aliastexcoord_t *st = NULL;

		if (SUCCEEDED (set->texcoords->lpVtbl->Lock (set->texcoords, 0, 0, (void **) &st, 0)))
		{
			int *order = (int *) ((byte *) hdr + hdr->ofs_glcmds);

			for (;;)
			{
				// get the vertex count and primitive type
				int count = *order++;

				if (!count) break;
				if (count < 0) count = -count;

				for (int i = 0; i < count; i++, st++, order += 3)
				{
					st->st[0] = ((float *) order)[0];
					st->st[1] = ((float *) order)[1];
				}
			}

			set->texcoords->lpVtbl->Unlock (set->texcoords);
		}
	}
}


void R_CreateIndexBuffer (dmdl_t *hdr, meshbuffers_t *set)
{
	if (SUCCEEDED (d3d_Device->lpVtbl->CreateIndexBuffer (d3d_Device, sizeof (uint16_t) * set->numindexes, D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED, &set->indexes, NULL)))
	{
		uint16_t *ndx = NULL;

		if (SUCCEEDED (set->indexes->lpVtbl->Lock (set->indexes, 0, 0, (void **) &ndx, 0)))
		{
			int *order = (int *) ((byte *) hdr + hdr->ofs_glcmds);
			int i, firstvertex = 0, numindexes = 0;

			for (;;)
			{
				// get the vertex count and primitive type
				int count = *order++;
				if (!count) break;

				if (count < 0)
				{
					for (i = 2, count = -count; i < count; i++, ndx += 3, numindexes += 3)
					{
						ndx[0] = firstvertex + 0;
						ndx[1] = firstvertex + (i - 1);
						ndx[2] = firstvertex + i;
					}
				}
				else
				{
					for (i = 2; i < count; i++, ndx += 3, numindexes += 3)
					{
						ndx[0] = firstvertex + i - 2;
						ndx[1] = firstvertex + ((i & 1) ? i : (i - 1));
						ndx[2] = firstvertex + ((i & 1) ? (i - 1) : i);
					}
				}

				order += count * 3;
				firstvertex += count;
			}

			set->indexes->lpVtbl->Unlock (set->indexes);

			assert (numindexes == set->numindexes);
		}
	}
}


void R_CountMeshBufferSizes (dmdl_t *hdr, meshbuffers_t *set)
{
	set->numverts = 0;
	set->numtris = 0;

	int *order = (int *) ((byte *) hdr + hdr->ofs_glcmds);

	for (;;)
	{
		// get the vertex count and primitive type
		int count = *order++;

		if (!count) break;
		if (count < 0) count = -count;

		order += (count * 3);

		set->numtris += (count - 2);
		set->numverts += count;
	}

	set->numindexes = set->numtris * 3;
}


void R_CreateMeshBuffers (model_t *mod, dmdl_t *hdr)
{
	// look for a match
	for (int i = 0; i < MAX_MESH_BUFFERS; i++)
	{
		// free slot
		if (!d3d_MeshBuffers[i].positions) continue;
		if (!d3d_MeshBuffers[i].texcoords) continue;
		if (!d3d_MeshBuffers[i].indexes) continue;

		if (!strcmp (mod->name, d3d_MeshBuffers[i].name))
		{
			// this is a match, register it, store to the model and return
			d3d_MeshBuffers[i].registration_sequence = r_registration_sequence;
			mod->bufferset = i;
			return;
		}
	}

	// look for a free buffer set
	for (int i = 0; i < MAX_MESH_BUFFERS; i++)
	{
		// not a free slot
		if (d3d_MeshBuffers[i].positions) continue;
		if (d3d_MeshBuffers[i].texcoords) continue;
		if (d3d_MeshBuffers[i].indexes) continue;

		// this is now a valid free slot so register it, store to the model, build the buffers and return
		strcpy (d3d_MeshBuffers[i].name, mod->name);
		d3d_MeshBuffers[i].registration_sequence = r_registration_sequence;
		mod->bufferset = i;

		// count of number of verts and indexes
		R_CountMeshBufferSizes (hdr, &d3d_MeshBuffers[i]);

		// build the buffers
		R_CreateFramesBuffer (hdr, &d3d_MeshBuffers[i]);
		R_CreateTexCoordsBuffer (hdr, &d3d_MeshBuffers[i]);
		R_CreateIndexBuffer (hdr, &d3d_MeshBuffers[i]);

		// done
		return;
	}

	// didn't find one
	ri->Sys_Error (ERR_DROP, "R_CreateMeshBuffers : failed to find a mesh buffer");
}


void R_RegisterMeshBuffers (model_t *mod)
{
	// ensure it's not thrown out
	d3d_MeshBuffers[mod->bufferset].registration_sequence = r_registration_sequence;
}


void R_FreeUnusedMeshBuffers (int rs)
{
	for (int i = 0; i < MAX_MESH_BUFFERS; i++)
	{
		if (d3d_MeshBuffers[i].registration_sequence != rs)
		{
			SAFE_RELEASE (d3d_MeshBuffers[i].positions);
			SAFE_RELEASE (d3d_MeshBuffers[i].texcoords);
			SAFE_RELEASE (d3d_MeshBuffers[i].indexes);

			memset (&d3d_MeshBuffers[i], 0, sizeof (d3d_MeshBuffers[i]));
		}
	}
}


/*
=============================================================

ALIAS MODELS

=============================================================
*/

image_t *R_GetAliasSkin (entity_t *e, model_t *mod)
{
	image_t	*skin = NULL;

	// select skin
	if (e->skin)
		skin = e->skin;	// custom player skin
	else
	{
		if (e->skinnum >= MAX_MD2SKINS)
			skin = mod->skins[0];
		else
		{
			skin = mod->skins[e->skinnum];

			if (!skin)
				skin = mod->skins[0];
		}
	}

	if (!skin)
		skin = r_notexture;	// fallback...

	return skin;
}


void R_DrawAliasPolyset (meshbuffers_t *set)
{
	// i may need to instance this stuff with lights going to per-instance data and passed in extra texcoord slots to the PS
	// not too sure what the gain would be - hypothetical, even though we have 64 dlights, mdls are commonly only affected by
	// 1 (muzzleflash, trail or explosion) and at most 2/3.  (confirmed - i went mad and measured it!!!)
	// an alternate approach might be to add in R_DynamicLightPoint dlights for small stuff (gibs/etc)
	// to be honest however it runs just fine so none of that is probably even needed
	d3d_Device->lpVtbl->DrawIndexedPrimitive (
		d3d_Device,
		D3DPT_TRIANGLELIST,
		0,
		0,
		set->numverts,
		0,
		set->numtris
		);

	c_draw_call++;

	// this is consistent with the original counts but because the entire MDL is
	// taken in a single draw call it's no longer representative of performance
	c_alias_polys += set->numtris;
}


/*
=============
GL_DrawAliasFrameLerp

interpolates between two frames and origins
=============
*/
void GL_DrawAliasFrameLerp (entity_t *e, model_t *mod)
{
	meshbuffers_t *set = &d3d_MeshBuffers[mod->bufferset];
	image_t	*skin = R_GetAliasSkin (e, mod);

	R_SetTexture (TS_DIFFUSE, skin);
	R_SetVertexDeclaration (d3d_MeshDecl);

	if (e->flags & (RF_SHELL_RED | RF_SHELL_GREEN | RF_SHELL_BLUE | RF_SHELL_DOUBLE | RF_SHELL_HALF_DAM))
	{
		R_SetVertexShader (d3d_MeshShellVS);
		R_SetPixelShader (d3d_MeshShellPS);
	}
	else
	{
		R_SetVertexShader (d3d_MeshTexturedVS);
		R_SetPixelShader (d3d_MeshTexturedPS);
	}

	R_SetVertexBuffer (0, set->positions, e->currframe * set->numverts * sizeof (aliaspolyvert_t), sizeof (aliaspolyvert_t));
	R_SetVertexBuffer (1, set->positions, e->prevframe * set->numverts * sizeof (aliaspolyvert_t), sizeof (aliaspolyvert_t));
	R_SetVertexBuffer (2, set->texcoords, 0, sizeof (aliastexcoord_t));
	R_SetIndexBuffer (set->indexes);

	R_DrawAliasPolyset (set);

	// no lights
	if (!R_ShouldEntityHaveDlights (e)) return;

	// add all the dynamic lights (to do - can we do this with instancing??? - probably not worth bothering cos
	// most of the time entities will only have a few lights on; might be a win when there's more though)
	for (int i = 0; i < r_newrefdef.num_dlights; i++)
	{
		dlight_t *dl = &r_newrefdef.dlights[i];
		float dist[3], add;

		VectorSubtract (e->currorigin, dl->origin, dist);
		add = dl->intensity - VectorLength (dist);

		if (add > 0)
		{
			// move the dlight back to entity space
			R_InverseTransform (&e->localmatrix, dl->transformed, dl->origin);

			// setup this light
			if (!R_BeginEntityLighting (e, dl)) continue;

			// all of our state is still up, so set some shaders and constants and proceed to draw it
			R_SetVertexShader (d3d_MeshDynamicVS);
			R_SetPixelShader (d3d_MeshDynamicPS);

			// draw the dlight pass
			R_DrawAliasPolyset (set);
		}
	}

	R_EndEntityLighting (e);
}


/*
===================
R_CullAliasModel
===================
*/
static qboolean R_CullAliasModel (vec3_t *bbox, entity_t *e)
{
	aliashdr_t *hdr = (aliashdr_t *) e->model->extradata;
	maliasframe_t *currframe = &hdr->frames[e->currframe];
	maliasframe_t *prevframe = &hdr->frames[e->prevframe];
	vec3_t mins, maxs;

	// compute axially aligned mins and maxs
	if (currframe == prevframe)
	{
		for (int i = 0; i < 3; i++)
		{
			mins[i] = currframe->translate[i];
			maxs[i] = mins[i] + currframe->scale[i] * 255;
		}
	}
	else
	{
		for (int i = 0; i < 3; i++)
		{
			float currmins = currframe->translate[i];
			float currmaxs = currmins + currframe->scale[i] * 255;

			float prevmins = prevframe->translate[i];
			float prevmaxs = prevmins + prevframe->scale[i] * 255;

			// lerp the box properly instead of just taking the smallest or greatest of mins or maxs
			mins[i] = prevmins + (currmins - prevmins) * (1.0f - e->backlerp);
			maxs[i] = prevmaxs + (currmaxs - prevmaxs) * (1.0f - e->backlerp);
		}
	}

	return R_CullEntityForBBox (e, bbox, mins, maxs);
}


void R_TransformAliasModel (entity_t *e, aliashdr_t *hdr)
{
	// set up the matrix transform
	R_SetEntityTransform (&e->localmatrix, e->flags);

	// set up the lerp transform
	float move[3], delta[3], vectors[3][3];
	float frontlerp = 1.0 - e->backlerp;

	// get current and previous frames
	maliasframe_t *currframe = &hdr->frames[e->currframe];
	maliasframe_t *prevframe = &hdr->frames[e->prevframe];

	// move should be the delta back to the previous frame * backlerp
	VectorSubtract (e->prevorigin, e->currorigin, delta);
	AngleVectors (e->angles, vectors[0], vectors[1], vectors[2]);

	move[0] = Vector3Dot (delta, vectors[0]);	// forward
	move[1] = -Vector3Dot (delta, vectors[1]);	// left
	move[2] = Vector3Dot (delta, vectors[2]);	// up

	VectorAdd (move, prevframe->translate, move);

	R_SetVertexShaderConstant3f (
		VSC_MESHMOVE,
		e->backlerp * move[0] + frontlerp * currframe->translate[0],
		e->backlerp * move[1] + frontlerp * currframe->translate[1],
		e->backlerp * move[2] + frontlerp * currframe->translate[2]
		);

	R_SetVertexShaderConstant3f (
		VSC_MESHFRONTV,
		frontlerp * currframe->scale[0],
		frontlerp * currframe->scale[1],
		frontlerp * currframe->scale[2]
		);

	R_SetVertexShaderConstant3f (
		VSC_MESHBACKV,
		e->backlerp * prevframe->scale[0],
		e->backlerp * prevframe->scale[1],
		e->backlerp * prevframe->scale[2]
		);

	if (!(e->flags & (RF_SHELL_RED | RF_SHELL_GREEN | RF_SHELL_BLUE | RF_SHELL_DOUBLE | RF_SHELL_HALF_DAM)))
		R_SetVertexShaderConstant1f (VSC_MESHSUITSCALE, 0);
	else R_SetVertexShaderConstant1f (VSC_MESHSUITSCALE, POWERSUIT_SCALE);

	R_SetVertexShaderConstant1f (VSC_MESHBLEND, e->backlerp);
}


void R_LightAliasModel (entity_t *e)
{
	float shadelight[4] = {0, 0, 0, 1};

	// clear the dynamic lighting flag
	e->flags &= ~RF_DYNAMICLIGHT;

	// get lighting information
	// PMM - rewrote, reordered to handle new shells & mixing
	if (e->flags & (RF_SHELL_HALF_DAM | RF_SHELL_GREEN | RF_SHELL_RED | RF_SHELL_BLUE | RF_SHELL_DOUBLE))
	{
		// PMM -special case for godmode
		if ((e->flags & RF_SHELL_RED) && (e->flags & RF_SHELL_BLUE) && (e->flags & RF_SHELL_GREEN))
		{
			shadelight[0] = 1.0;
			shadelight[1] = 1.0;
			shadelight[2] = 1.0;
		}
		else if (e->flags & (RF_SHELL_RED | RF_SHELL_BLUE | RF_SHELL_DOUBLE))
		{
			VectorClear (shadelight);

			if (e->flags & RF_SHELL_RED)
			{
				shadelight[0] = 1.0;

				if (e->flags & (RF_SHELL_BLUE | RF_SHELL_DOUBLE))
					shadelight[2] = 1.0;
			}
			else if (e->flags & RF_SHELL_BLUE)
			{
				if (e->flags & RF_SHELL_DOUBLE)
				{
					shadelight[1] = 1.0;
					shadelight[2] = 1.0;
				}
				else shadelight[2] = 1.0;
			}
			else if (e->flags & RF_SHELL_DOUBLE)
			{
				shadelight[0] = 0.9;
				shadelight[1] = 0.7;
			}
		}
		else if (e->flags & (RF_SHELL_HALF_DAM | RF_SHELL_GREEN))
		{
			VectorClear (shadelight);

			// PMM - new colors
			if (e->flags & RF_SHELL_HALF_DAM)
			{
				shadelight[0] = 0.56;
				shadelight[1] = 0.59;
				shadelight[2] = 0.45;
			}

			if (e->flags & RF_SHELL_GREEN)
				shadelight[1] = 1.0;
		}
	}
	else if (e->flags & RF_FULLBRIGHT)
	{
		shadelight[0] = 1.0;
		shadelight[1] = 1.0;
		shadelight[2] = 1.0;
	}
	else
	{
		// don't add dynamics here because we'll blend them in using extra passes later on
		// doing this allows large models (such as flags) recieve dlights properly rather than just being fully lit
		// (null models and R_SetLightLevel still use R_DynamicLightPoint)
		R_LightPoint (e->currorigin, shadelight);

		// this entity now gets dynamic lights
		e->flags |= RF_DYNAMICLIGHT;

		// player lighting level was fixed in R_SetLightLevel so it's no longer necessary to do it here
		// it never worked right in R_SetLightLevel because the original code used "currententity" all
		// over the place, which was horrible.
	}

	if (e->flags & RF_MINLIGHT)
	{
		float scale = 8.0f / 255.0f;

		shadelight[0] = (shadelight[0] + scale) / (1.0f + scale);
		shadelight[1] = (shadelight[1] + scale) / (1.0f + scale);
		shadelight[2] = (shadelight[2] + scale) / (1.0f + scale);
	}

	if (e->flags & RF_GLOW)
	{
		// bonus items will pulse with time
		float scale = 0.1 * sin (r_newrefdef.time * 7);

		for (int i = 0; i < 3; i++)
		{
			float min = shadelight[i] * 0.8;

			shadelight[i] += scale;

			if (shadelight[i] < min)
				shadelight[i] = min;
		}
	}

	R_SetEntityLighting (e, shadelight);
}


/*
=================
R_DrawAliasModel

=================
*/
void R_DrawAliasModel (entity_t *e)
{
	int			i;
	aliashdr_t	*hdr = (aliashdr_t *) e->model->extradata;
	vec3_t		bbox[8];

	if ((e->currframe >= hdr->numframes) || (e->currframe < 0))
	{
		ri->Con_Printf (PRINT_ALL, "R_DrawAliasModel %s: no such currframe %d\n", e->model->name, e->currframe);
		e->currframe = 0;
		e->prevframe = 0;
	}

	if ((e->prevframe >= hdr->numframes) || (e->prevframe < 0))
	{
		ri->Con_Printf (PRINT_ALL, "R_DrawAliasModel %s: no such prevframe %d\n", e->model->name, e->prevframe);
		e->currframe = 0;
		e->prevframe = 0;
	}

	if (!(e->flags & RF_WEAPONMODEL))
	{
		if (R_CullAliasModel (bbox, e))
			return;
	}
	else
	{
		if (r_lefthand->value == 2)
			return;
	}

	// set up state for the alias entity
	R_SetEntityState (e);

	if (!r_lerpmodels->value)
		e->backlerp = 0;

	// draw all the triangles
	R_LightAliasModel (e);
	R_TransformAliasModel (e, hdr);
	GL_DrawAliasFrameLerp (e, e->model);

	// draw a bounding box for the entity if it's not the weapon model
	if (!(e->flags & RF_WEAPONMODEL) && r_showbboxes->value) R_DrawBBoxBox (bbox);
}

