/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_main.c

#include "gl_local.h"
#include "resource.h"

typedef struct spritebuffer_s
{
	IDirect3DVertexBuffer9 *VertexBuffer;

	char name[MAX_PATH];
	int registration_sequence;
} spritebuffer_t;

#define MAX_SPRITE_BUFFERS	1024

spritebuffer_t d3d_SpriteBuffers[MAX_SPRITE_BUFFERS];

IDirect3DVertexShader9 *d3d_SpriteVS = NULL;
IDirect3DPixelShader9 *d3d_SpritePS = NULL;
IDirect3DVertexDeclaration9 *d3d_SpriteDecl = NULL;

typedef struct spritepolyvert_s
{
	float xy[2];
	float st[2];
} spritepolyvert_t;


void Sprite_OnCreate (void)
{
	D3DVERTEXELEMENT9 layout[] = {
		VDECL (0, offsetof (spritepolyvert_t, xy), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_POSITION, 0),
		VDECL (0, offsetof (spritepolyvert_t, st), D3DDECLTYPE_FLOAT2, D3DDECLUSAGE_TEXCOORD, 0),
		D3DDECL_END ()
	};

	d3d_SpriteDecl = R_CreateVertexDeclaration (layout);

	R_CreateShaders (SpriteVS, SpritePS, IDR_SPRITESHADER);
}


void Sprite_OnReset (void) {}
void Sprite_OnLost (void) {}

void Sprite_OnDestroy (void)
{
	for (int i = 0; i < MAX_SPRITE_BUFFERS; i++)
	{
		SAFE_RELEASE (d3d_SpriteBuffers[i].VertexBuffer);
		memset (&d3d_SpriteBuffers[i], 0, sizeof (spritebuffer_t));
	}
}

void Sprite_Register (void)
{
	static qboolean registered = false;

	if (!registered)
	{
		R_RegisterHandler (Sprite_OnCreate, Sprite_OnLost, Sprite_OnReset, Sprite_OnDestroy);
		registered = true;
	}
}


void R_CreateSpriteBuffer (dsprite_t *hdr, spritebuffer_t *set)
{
	spritepolyvert_t *verts = NULL;

	d3d_Device->lpVtbl->CreateVertexBuffer (d3d_Device, hdr->numframes * sizeof (spritepolyvert_t) * 4, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &set->VertexBuffer, NULL);

	set->VertexBuffer->lpVtbl->Lock (set->VertexBuffer, 0, 0, (void **) &verts, 0);

	for (int i = 0; i < hdr->numframes; i++, verts += 4)
	{
		dsprframe_t	*frame = &hdr->frames[i];

		verts[0].xy[0] = -frame->origin_x;
		verts[0].xy[1] = -frame->origin_y;
		verts[0].st[0] = 0;
		verts[0].st[1] = 1;

		verts[1].xy[0] = -frame->origin_x;
		verts[1].xy[1] = frame->height - frame->origin_y;
		verts[1].st[0] = 0;
		verts[1].st[1] = 0;

		verts[2].xy[0] = frame->width - frame->origin_x;
		verts[2].xy[1] = -frame->origin_y;
		verts[2].st[0] = 1;
		verts[2].st[1] = 1;

		verts[3].xy[0] = frame->width - frame->origin_x;
		verts[3].xy[1] = frame->height - frame->origin_y;
		verts[3].st[0] = 1;
		verts[3].st[1] = 0;
	}

	set->VertexBuffer->lpVtbl->Unlock (set->VertexBuffer);
}


void R_CreateSpriteBuffers (model_t *mod)
{
	// look for a match
	for (int i = 0; i < MAX_SPRITE_BUFFERS; i++)
	{
		// free slot
		if (!d3d_SpriteBuffers[i].VertexBuffer) continue;

		if (!strcmp (mod->name, d3d_SpriteBuffers[i].name))
		{
			// this is a match, register it, store to the model and return
			d3d_SpriteBuffers[i].registration_sequence = r_registration_sequence;
			mod->bufferset = i;
			return;
		}
	}

	// look for a free buffer set
	for (int i = 0; i < MAX_SPRITE_BUFFERS; i++)
	{
		// not a free slot
		if (d3d_SpriteBuffers[i].VertexBuffer) continue;

		// this is now a valid free slot so register it, store to the model, build the buffers and return
		strcpy (d3d_SpriteBuffers[i].name, mod->name);
		d3d_SpriteBuffers[i].registration_sequence = r_registration_sequence;
		mod->bufferset = i;

		// build the buffers
		R_CreateSpriteBuffer ((dsprite_t *) mod->extradata, &d3d_SpriteBuffers[i]);

		// done
		return;
	}

	// didn't find one
	ri->Sys_Error (ERR_DROP, "R_CreateSpriteBuffers : failed to find a Sprite buffer");
}


void R_RegisterSpriteBuffers (model_t *mod)
{
	d3d_SpriteBuffers[mod->bufferset].registration_sequence = r_registration_sequence;
}


void R_FreeUnusedSpriteBuffers (int rs)
{
	for (int i = 0; i < MAX_SPRITE_BUFFERS; i++)
	{
		if (d3d_SpriteBuffers[i].registration_sequence != rs)
		{
			SAFE_RELEASE (d3d_SpriteBuffers[i].VertexBuffer);
			memset (&d3d_SpriteBuffers[i], 0, sizeof (d3d_SpriteBuffers[i]));
		}
	}
}


/*
=================
R_DrawSpriteModel

=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	dsprframe_t	*frame;

	// don't even bother culling, because it's just a single polygon without a surface cache
	dsprite_t *psprite = (dsprite_t *) e->model->extradata;

	// get the frame to draw
	e->currframe %= psprite->numframes;
	frame = &psprite->frames[e->currframe];

	// set up state for the sprite entity
	R_SetRenderStates (RS_DEPTHTEST | RS_CULLCCW | RS_ALPHABLEND | RS_ALPHATEST);

	// shaders
	R_SetVertexShader (d3d_SpriteVS);
	R_SetPixelShader (d3d_SpritePS);

	// constants - entity origin is a struct member and can be safely overflowed; the other vars are padded
	R_SetVertexShaderConstant4fv (VSC_SPRITEORIGIN, e->currorigin, 1);

	// texture and decl
	R_SetTexture (TS_DIFFUSE, e->model->skins[e->currframe]);

	R_SetVertexDeclaration (d3d_SpriteDecl);
	R_SetVertexBuffer (0, d3d_SpriteBuffers[e->model->bufferset].VertexBuffer, 0, sizeof (spritepolyvert_t));

	// and now draw it
	d3d_Device->lpVtbl->DrawPrimitive (d3d_Device, D3DPT_TRIANGLESTRIP, e->currframe * 4, 2);
	c_draw_call++;
}


