/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "gl_local.h"

/*
================
R_RecursiveWorldNode
================
*/
int cut_area_leafs = 0;

void R_RecursiveWorldNode (refdef_t *rd, entity_t *e, mnode_t *node, int clipflags)
{
	if (node->contents == CONTENTS_SOLID) return;
	if (node->visframe != r_visframecount) return;

	if (clipflags)
	{
		for (int c = 0; c < 4; c++)
		{
			// don't need to clip against it
			if (!(clipflags & (1 << c))) continue;

			int side = BoxOnPlaneSide (node->mins, node->maxs, &frustum[c]);

			if (side == 1) clipflags &= ~(1 << c);	// node is entirely on screen
			if (side == 2) return;	// node is entirely off screen
		}
	}

	// if a leaf node, draw stuff
	if (node->contents != -1)
	{
		mleaf_t *pleaf = (mleaf_t *) node;

		// check for door connected areas
		if (r_newrefdef.areabits)
		{
			if (!(r_newrefdef.areabits[pleaf->area >> 3] & (1 << (pleaf->area & 7))))
			{
				// not visible
				cut_area_leafs++;
				return;
			}
		}

		msurface_t **mark = pleaf->firstmarksurface;
		int c = pleaf->nummarksurfaces;

		if (c)
		{
			do
			{
				(*mark)->visframe = r_framecount;
				mark++;
			} while (--c);
		}

		return;
	}

	// node is just a decision point, so go down the apropriate sides
	cplane_t *plane = node->plane;
	float dot = Vector3Dot (e->modelorg, plane->normal) - plane->dist;

	// find which side of the node we are on
	int side = (dot >= 0) ? 0 : 1;
	int sidebit = (dot >= 0) ? 0 : SURF_PLANEBACK;

	// recurse down the children, front side first
	R_RecursiveWorldNode (rd, e, node->children[side], clipflags);

	for (int c = 0; c < node->numsurfaces; c++)
	{
		// draw stuff
		msurface_t *surf = &node->surfaces[c];

		// not visible (surf->dlightframe controls whether the surf was drawn so we don't need to hack the visframe to exclude
		// surfs culled in the following 2 steps here)
		if (surf->visframe != r_framecount) continue;

		// now cull backfacing surfs
		if ((surf->flags & SURF_PLANEBACK) != sidebit) continue;

		// this is worth a few extra frames; nothing dramatic but it's so easy that we may as well
		if (R_CullBox (surf->mins, surf->maxs)) continue;

		// and chain it up
		R_ChainSurface (e, surf);
	}

	// recurse down the back side
	R_RecursiveWorldNode (rd, e, node->children[!side], clipflags);
}


/*
=============
R_DrawWorld
=============
*/
void R_DrawWorld (refdef_t *rd, entity_t *r_worldentity)
{
	if (!r_drawworld->value) return;
	if (rd->rdflags & RDF_NOWORLDMODEL) return;

	// auto cycle the world frame for texture animation
	memset (r_worldentity, 0, sizeof (entity_t));

	Vector3Copy (r_worldentity->modelorg, rd->vieworigin);

	// set up other properties of the entity
	r_worldentity->currframe = (int) (rd->time * 2);
	r_worldentity->model = r_worldmodel;

	// this is a hack to ensure that the world always draws, otherwise it may not if a door (for example)
	// takes the entire view but it's areaportal is closed, so the only surfs are merged from that door.
	// in general that's the only condition in which R_RecursiveWorldNode may add no surfs, but even then
	// the merged door will add it's surfs, so it's valid
	r_worldentity->numdrawsurfs = 1;

	// clear everything down before drawing
	R_NewImageFrame ();
	R_ClearTextureChains ();

	// the world is not transformed
	R_IdentityMatrix (&r_worldentity->localmatrix);
	r_worldentity->translated = false;
	r_worldentity->rotated = false;

	// add world surfaces to the draw list
	cut_area_leafs = 0;
	R_RecursiveWorldNode (rd, r_worldentity, r_worldmodel->nodes, 15);

	// and draw the lot
	R_DrawTextureChains (r_worldentity);
}


void R_AddViewLeaf (mleaf_t *leaf)
{
	if (!leaf) return;

	byte *vis = Mod_ClusterPVS (leaf->cluster, r_worldmodel);

	for (int i = 0; i < r_worldmodel->numleafs; i++)
	{
		mleaf_t	*leaf = &r_worldmodel->leafs[i];
		int cluster = leaf->cluster;

		if (cluster == -1)
			continue;

		if (vis[cluster >> 3] & (1 << (cluster & 7)))
		{
			mnode_t *node = (mnode_t *) leaf;

			do
			{
				if (node->visframe == r_visframecount)
					break;

				node->visframe = r_visframecount;
				node = node->parent;
			} while (node);
		}
	}
}


/*
===============
R_MarkLeaves

Mark the leaves and nodes that are in the PVS for the current cluster
===============
*/
void R_MarkLeaves (refdef_t *rd)
{
	if (gl_lockpvs->modified || r_novis->modified)
	{
		// NULL out the old so that the PVS will recalc
		r_viewleaf[1] = NULL;

		// no longer modified now
		gl_lockpvs->modified = false;
		r_novis->modified = false;
	}

	// r_worldmodel is not set
	if (rd->rdflags & RDF_NOWORLDMODEL) return;

	// PVS is unchanged
	if (r_viewleaf[0] && r_viewleaf[1] && r_viewleaf[0]->cluster == r_viewleaf[1]->cluster) return;

	// go to a new PVS frame
	r_visframecount++;

	if (r_novis->value || !r_viewleaf[0] || !r_worldmodel->vis)
	{
		// mark everything
		for (int i = 0; i < r_worldmodel->numleafs; i++) r_worldmodel->leafs[i].visframe = r_visframecount;
		for (int i = 0; i < r_worldmodel->numnodes; i++) r_worldmodel->nodes[i].visframe = r_visframecount;
	}
	else
	{
		// add the main viewleaf
		R_AddViewLeaf (r_viewleaf[0]);

		// check for a contents transition
		if (r_viewleaf[1] && r_viewleaf[1]->contents != r_viewleaf[0]->contents)
			R_AddViewLeaf (r_viewleaf[1]);
	}

	// set back
	r_viewleaf[1] = r_viewleaf[0];
}

