/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// common.c -- misc functions used in client and server

#include "quakedef.h"


//============================================================================

// 64mb ought to be enough for anybody :)
// this was 10 but some of the boss models overflow it...
byte scratchbuf[64 * 1024 * 1024];
int scratch_lowmark = 0;

void *Scratch_Alloc (void)
{
	// allocates without advancing the lowmark
	return &scratchbuf[scratch_lowmark];
}


void *Scratch_AllocSize (int size)
{
	// allocates from scratch advancing the lowmark
	byte *buf = &scratchbuf[scratch_lowmark];

	scratch_lowmark += size;

	return buf;
}


void Scratch_FreeToLowMark (int mark)
{
	// just resets the lowmark
	scratch_lowmark = mark;
}


int Scratch_LowMark (void)
{
	return scratch_lowmark;
}


//===============================================================================

int		hunkcount;

byte	*membase;
int		hunkmaxsize;
int		cursize;


void *Hunk_Begin (int maxsize)
{
	// reserve a huge chunk of memory, but don't commit any yet
	cursize = 0;
	hunkmaxsize = maxsize;
	membase = VirtualAlloc (NULL, maxsize, MEM_RESERVE, PAGE_NOACCESS);

	if (!membase)
		Sys_Error ("VirtualAlloc reserve failed");

	return (void *) membase;
}


void *Hunk_Alloc (int size)
{
	// round to cacheline
	size = (size + 31) & ~31;

	// commit pages as needed
	void *buf = VirtualAlloc (&membase[cursize], size, MEM_COMMIT, PAGE_READWRITE);

	if (!buf)
	{
		FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError (), MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT), (LPTSTR) &buf, 0, NULL);
		Sys_Error ("VirtualAlloc commit failed.\n%s", buf);
		return NULL;
	}

	buf = (void *) &membase[cursize];
	cursize += size;

	if (cursize > hunkmaxsize)
		Sys_Error ("Hunk_Alloc overflow");

	return (void *) buf;
}


int Hunk_End (void)
{
	hunkcount++;
	return cursize;
}


void Hunk_Free (void *base)
{
	if (base)
		VirtualFree (base, 0, MEM_RELEASE);

	hunkcount--;
}


/*
==============================================================================

ZONE MEMORY ALLOCATION

using heap allocations to prevent a crash from a bad Z_Free when calling gi.FreeTags
via ge->SpawnEntities during loading of eou7.cin (Z_FreeTags now just throws away
the full heap rather than using individual Z_Free calls which makes it safe)

==============================================================================
*/

// it's possible for games to exceed this (neither baseq2 nor ctf do) which would cause an error meaning we need to increase it
#define MAX_ZONETAGS	1024
#define Z_MAGIC			0x35012560

typedef struct z_alloc_s
{
	int size;
	int tag;
	int magic;
} z_alloc_t;


typedef struct zone_s
{
	HANDLE hZone;
	int tag;
	int count;
	int bytes;
} zone_t;

zone_t z_zones[MAX_ZONETAGS];

/*
========================
Z_Free
========================
*/
void Z_Free (void *ptr)
{
	// guard against double-freeing
	if (!ptr) return;

	// retrieve the allocation
	z_alloc_t *zz = ((z_alloc_t *) ptr) - 1;

	// sanity-check it so that nothing is incorrectly freed; if it's mismatched we just leak it and the next Z_FreeTags will clean it up
	if (zz->magic != Z_MAGIC) return;
	if (zz->tag < 0) return;
	if (zz->tag >= MAX_ZONETAGS) return;

	// retrieve the zone
	zone_t *z = &z_zones[zz->tag];

	// making sure that it's valid
	if (!z->hZone) return;

	// counts
	z->bytes -= zz->size;
	z->count--;

	// and free it properly, then compact the heap to prevent memory fragmentation
	HeapFree (z->hZone, 0, zz);
	HeapCompact (z->hZone, 0);
}


/*
========================
Z_Stats_f
========================
*/
void Z_Stats_f (void)
{
	int z_bytes = 0;
	int z_count = 0;

	for (int i = 0; i < MAX_ZONETAGS; i++)
	{
		z_bytes += z_zones[i].bytes;
		z_count += z_zones[i].count;
	}

	Com_Printf ("%i bytes in %i blocks\n", z_bytes, z_count);
}


/*
========================
Z_FreeTags
========================
*/
void Z_FreeTags (int tag)
{
	// this should never happen; if it does then we need to know
	if (tag < 0) Com_Error (ERR_FATAL, "Z_FreeTags: bad tag");
	if (tag >= MAX_ZONETAGS) Com_Error (ERR_FATAL, "Z_FreeTags: bad tag");

	// destroy and fully clear the zone
	HeapDestroy (z_zones[tag].hZone);
	memset (&z_zones[tag], 0, sizeof (z_zones[tag]));
}


/*
========================
Z_TagAlloc
========================
*/
void *Z_TagAlloc (int size, int tag)
{
	if (tag < 0) Com_Error (ERR_FATAL, "Z_TagAlloc: bad tag");
	if (tag >= MAX_ZONETAGS) Com_Error (ERR_FATAL, "Z_TagAlloc: bad tag");

	zone_t *z = &z_zones[tag];

	// create the heap if necessary
	if (!z->hZone) z->hZone = HeapCreate (0, 0, 0);
	if (!z->hZone) Com_Error (ERR_FATAL, "Z_Alloc: failed on HeapCreate");

	// make the zone block allocation
	z_alloc_t *zz = (z_alloc_t *) HeapAlloc (z->hZone, HEAP_ZERO_MEMORY, size + sizeof (z_alloc_t));

	// fill it in
	zz->magic = Z_MAGIC;
	zz->tag = tag;
	zz->size = size;

	// counts
	z->bytes += size;
	z->count++;

	// return what we got
	return (void *) (zz + 1);
}


/*
========================
Z_Alloc
========================
*/
void *Z_Alloc (int size)
{
	return Z_TagAlloc (size, 0);
}


void Z_Init (void)
{
	memset (z_zones, 0, sizeof (z_zones));
	Cmd_AddCommand ("z_stats", Z_Stats_f);
}


//============================================================================


void SZ_Init (sizebuf_t *buf, byte *data, int length)
{
	memset (buf, 0, sizeof (*buf));
	buf->data = data;
	buf->maxsize = length;
}

void SZ_Clear (sizebuf_t *buf)
{
	buf->cursize = 0;
	buf->overflowed = false;
}

void *SZ_GetSpace (sizebuf_t *buf, int length)
{
	void *data;

	if (buf->cursize + length > buf->maxsize)
	{
		if (!buf->allowoverflow) Com_Error (ERR_FATAL, "SZ_GetSpace: overflow without allowoverflow set");
		if (length > buf->maxsize) Com_Error (ERR_FATAL, "SZ_GetSpace: %i is > full buffer size", length);

		Com_Printf ("SZ_GetSpace: overflow\n");

		SZ_Clear (buf);
		buf->overflowed = true;
	}

	data = buf->data + buf->cursize;
	buf->cursize += length;

	return data;
}

void SZ_Write (sizebuf_t *buf, void *data, int length)
{
	memcpy (SZ_GetSpace (buf, length), data, length);
}

void SZ_Print (sizebuf_t *buf, char *data)
{
	int		len;

	len = strlen (data) + 1;

	if (buf->cursize)
	{
		if (buf->data[buf->cursize - 1])
			memcpy ((byte *) SZ_GetSpace (buf, len), data, len); // no trailing 0
		else
			memcpy ((byte *) SZ_GetSpace (buf, len - 1) - 1, data, len); // write over trailing 0
	}
	else
		memcpy ((byte *) SZ_GetSpace (buf, len), data, len);
}


void FreeItemList (itemlist_t *list)
{
	if (list->heap)
	{
		HeapDestroy (list->heap);
		list->heap = NULL;
	}

	list->numitems = 0;
}


void BuildItemList (itemlist_t *list, void (*BuildFunc) (itemlist_t *, void *), void *extradata, qboolean sort)
{
	// take an actual scratch buffer so that we can safely Scratch_Alloc within the build func
	int mark = Scratch_LowMark ();
	char **items = (char **) Scratch_AllocSize (MAX_LIST_ITEMS * sizeof (char *));

	// clear down anything that may already exist
	FreeItemList (list);

	// create a new heap and a temporary pointer for the items
	list->heap = HeapCreate (0, 0, 0);
	list->items = items;

	// call the builder on the list
	BuildFunc (list, extradata);

	// allocate it for real and copy them over, adding an empty item at the end in case we need to do null termination checking
	list->items = (char **) HeapAlloc (list->heap, HEAP_ZERO_MEMORY, (list->numitems + 1) * sizeof (char *));
	memcpy (list->items, items, list->numitems * sizeof (char *));

	// optionally sort the list

	// free scratch buffer
	Scratch_FreeToLowMark (mark);
}


void CopyItemListItem (itemlist_t *list, char *src)
{
	if (list->numitems < MAX_LIST_ITEMS)
	{
		// copy over a new item to the list
		list->items[list->numitems] = (char *) HeapAlloc (list->heap, HEAP_ZERO_MEMORY, strlen (src) + 1);
		strcpy (list->items[list->numitems], src);

		list->numitems++;
	}
}


