/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// Main windowed and fullscreen graphics interface module. This module
// is used for both the software and OpenGL rendering versions of the
// Quake refresh engine.

#include "quakedef.h"

// needed for PrepVideoMenu
#include <d3d9.h>

// if this is changed it must be changed in glw_imp.c too (at least until we complete the transition)
#define	WINDOW_CLASS_NAME	"DirectQIIRedux"

void M_PrepVideoMenu (D3DDISPLAYMODE *modes, int nummodes);

// Structure containing functions exported from refresh DLL
refexport_t	*re;

// video cvars accessible from ri and re
vidvars_t vidvars;

// Global variables used internally by this module
viddef_t	viddef;				// global video state; used by other modules
qboolean reflib_active = 0;

HWND        cl_hwnd;            // Main window handle for life of program

#define VID_NUM_MODES (sizeof (vid_modes) / sizeof (vid_modes[0]))

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

static qboolean s_alttab_disabled;

extern	unsigned	sys_msg_time;

/*
==========================================================================

DLL GLUE

==========================================================================
*/

#define	MAXPRINTMSG	4096

void VID_Printf (int print_level, char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];
	static qboolean inupdate;

	va_start (argptr, fmt);
	vsprintf (msg, fmt, argptr);
	va_end (argptr);

	if (print_level == PRINT_ALL)
	{
		Com_Printf ("%s", msg);
	}
	else if (print_level == PRINT_DEVELOPER)
	{
		Com_DPrintf ("%s", msg);
	}
	else if (print_level == PRINT_ALERT)
	{
		MessageBox (0, msg, "PRINT_ALERT", MB_ICONWARNING);
		OutputDebugString (msg);
	}
}

void VID_Error (int err_level, char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];
	static qboolean inupdate;

	va_start (argptr, fmt);
	vsprintf (msg, fmt, argptr);
	va_end (argptr);

	Com_Error (err_level, "%s", msg);
}

//==========================================================================

byte        scantokey[128] =
{
	//  0           1       2       3       4       5       6       7
	//  8           9       A       B       C       D       E       F
	0, 27, '1', '2', '3', '4', '5', '6',
	'7', '8', '9', '0', '-', '=', K_BACKSPACE, 9, // 0
	'q', 'w', 'e', 'r', 't', 'y', 'u', 'i',
	'o', 'p', '[', ']', 13, K_CTRL, 'a', 's',     // 1
	'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',
	'\'', '`', K_SHIFT, '\\', 'z', 'x', 'c', 'v',     // 2
	'b', 'n', 'm', ',', '.', '/', K_SHIFT, '*',
	K_ALT, ' ', 0, K_F1, K_F2, K_F3, K_F4, K_F5,  // 3
	K_F6, K_F7, K_F8, K_F9, K_F10, K_PAUSE, 0, K_HOME,
	K_UPARROW, K_PGUP, K_KP_MINUS, K_LEFTARROW, K_KP_5, K_RIGHTARROW, K_KP_PLUS, K_END, //4
	K_DOWNARROW, K_PGDN, K_INS, K_DEL, 0, 0, 0, K_F11,
	K_F12, 0, 0, 0, 0, 0, 0, 0,       // 5
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0,        // 6
	0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0         // 7
};

/*
=======
MapKey

Map from windows to quake keynums
=======
*/
int MapKey (int key)
{
	int result;
	int modified = (key >> 16) & 255;
	qboolean is_extended = false;

	if (modified > 127)
		return 0;

	if (key & (1 << 24))
		is_extended = true;

	result = scantokey[modified];

	if (!is_extended)
	{
		switch (result)
		{
		case K_HOME:
			return K_KP_HOME;
		case K_UPARROW:
			return K_KP_UPARROW;
		case K_PGUP:
			return K_KP_PGUP;
		case K_LEFTARROW:
			return K_KP_LEFTARROW;
		case K_RIGHTARROW:
			return K_KP_RIGHTARROW;
		case K_END:
			return K_KP_END;
		case K_DOWNARROW:
			return K_KP_DOWNARROW;
		case K_PGDN:
			return K_KP_PGDN;
		case K_INS:
			return K_KP_INS;
		case K_DEL:
			return K_KP_DEL;
		default:
			return result;
		}
	}
	else
	{
		switch (result)
		{
		case 0x0D:
			return K_KP_ENTER;
		case 0x2F:
			return K_KP_SLASH;
		case 0xAF:
			return K_KP_PLUS;
		}

		return result;
	}
}

void AppActivate (BOOL fActive, BOOL minimize)
{
	Minimized = minimize;

	Key_ClearStates ();

	// we don't want to act like we're active if we're minimized
	if (fActive && !Minimized)
		ActiveApp = true;
	else ActiveApp = false;

	// minimize/restore mouse-capture on demand
	if (!ActiveApp)
	{
		IN_Activate (false);
		S_Activate (false);
	}
	else
	{
		IN_Activate (true);
		S_Activate (true);
	}
}

/*
====================
MainWndProc

main window procedure
====================
*/
void VID_HandleWMActivate (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int	fActive, fMinimized;

	// KJB: Watch this for problems in fullscreen modes with Alt-tabbing.
	fActive = LOWORD (wParam);
	fMinimized = (BOOL) HIWORD (wParam);

	AppActivate (fActive != WA_INACTIVE, fMinimized);

	if (reflib_active)
		re->AppActivate (!(fActive == WA_INACTIVE));
}


void VID_HandleWMMove (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (!vidvars.fullscreen->value)
	{
		int xPos = (short) LOWORD (lParam);   // horizontal position
		int yPos = (short) HIWORD (lParam);   // vertical position

		// right and bottom can be anything cos we just need the adjusted left and top here
		RECT r = {0, 0, 1, 1};

		AdjustWindowRect (&r, GetWindowLong (hWnd, GWL_STYLE), FALSE);

		Cvar_SetValue ("vid_xpos", xPos + r.left);
		Cvar_SetValue ("vid_ypos", yPos + r.top);

		if (ActiveApp)
			IN_Activate (true);
	}
}


void VID_HandleExitSizeMove (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if (!vidvars.fullscreen->value)
	{
		RECT crect;

		GetClientRect (cl_hwnd, &crect);
		Cvar_SetValue ("vid_width", crect.right - crect.left);
		Cvar_SetValue ("vid_height", crect.bottom - crect.top);
	}
}


LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_EXITSIZEMOVE:
		VID_HandleExitSizeMove (hWnd, uMsg, wParam, lParam);
		return 0;

	case WM_ENTERSIZEMOVE:
	case WM_ENTERMENULOOP:
		// deactivate the mouse
		IN_Activate (false);

		// fall through to DefWindowProc
		break;

	case WM_ERASEBKGND:
		// per MSDN this should return non-zero if we handle it ourselves
		return 1;

	case WM_MOUSEWHEEL:
		// this chunk of code theoretically only works under NT4 and Win98 since this message doesn't exist under Win95
		if ((short) HIWORD (wParam) > 0)
		{
			Key_Event (K_MWHEELUP, true, sys_msg_time);
			Key_Event (K_MWHEELUP, false, sys_msg_time);
		}
		else
		{
			Key_Event (K_MWHEELDOWN, true, sys_msg_time);
			Key_Event (K_MWHEELDOWN, false, sys_msg_time);
		}

		return 0;

	case WM_HOTKEY:
		return 0;

	case WM_CREATE:
		cl_hwnd = hWnd;
		return 0;

	case WM_DESTROY:
		// let sound and input know about this?
		cl_hwnd = NULL;
		return 0;

	case WM_ACTIVATE:
		VID_HandleWMActivate (hWnd, uMsg, wParam, lParam);
		return 0;

	case WM_MOVE:
		// break, not return 0, because we want to allow the move
		VID_HandleWMMove (hWnd, uMsg, wParam, lParam);
		break;

		// this is complicated because Win32 seems to pack multiple mouse events into
		// one update sometimes, so we always check all states and look for events
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_MBUTTONDOWN:
		if (wParam & MK_LBUTTON) Key_Event (K_MOUSE1, true, sys_msg_time);
		if (wParam & MK_RBUTTON) Key_Event (K_MOUSE2, true, sys_msg_time);
		if (wParam & MK_MBUTTON) Key_Event (K_MOUSE3, true, sys_msg_time);
		return 0;

	case WM_LBUTTONUP:
	case WM_RBUTTONUP:
	case WM_MBUTTONUP:
		if (wParam & MK_LBUTTON) Key_Event (K_MOUSE1, false, sys_msg_time);
		if (wParam & MK_RBUTTON) Key_Event (K_MOUSE2, false, sys_msg_time);
		if (wParam & MK_MBUTTON) Key_Event (K_MOUSE3, false, sys_msg_time);
		return 0;

	case WM_SYSCOMMAND:
		// prevent the screensaver from running but allow all other commands
		// (d3d9 disables screensavers automatically, but only in fs modes, and we may wish to cvar-ize this option)
		if (wParam == SC_SCREENSAVE)
			return 0;
		else break;

	case WM_SYSKEYDOWN:
		// vidvars.fullscreen, not vidvars.fullscreen->value, is intentional here as it's checking that the cvar exists
		if (wParam == 13 && vidvars.fullscreen)
		{
			// Alt-Enter
			re->HandleAltEnter ();
			return 0;
		}

		// fall through
	case WM_KEYDOWN:
		Key_Event (MapKey (lParam), true, sys_msg_time);
		return 0;

	case WM_SYSKEYUP:
	case WM_KEYUP:
		Key_Event (MapKey (lParam), false, sys_msg_time);
		return 0;

	default:
		break;
	}

	// pass all unhandled messages to DefWindowProc
	return DefWindowProc (hWnd, uMsg, wParam, lParam);
}


/*
============
VID_Restart_f

Console command to re-start the video mode and refresh DLL. We do this
simply by setting the modified flag for the vid variables, which will
cause the entire video mode and refresh DLL to be reset on the next frame.
============
*/
void VID_Restart_f (void)
{
	re->ForceVidRestart (re->vars);
}


void VID_CenterWindow_f (void)
{
	RECT windowrect;

	if (GetWindowRect (cl_hwnd, &windowrect))
	{
		RECT workarea;

		if (SystemParametersInfo (SPI_GETWORKAREA, 0, &workarea, 0))
		{
			Cvar_SetValue ("vid_xpos", workarea.left + ((workarea.right - workarea.left) - (windowrect.right - windowrect.left)) / 2);
			Cvar_SetValue ("vid_ypos", workarea.top + ((workarea.bottom - workarea.top) - (windowrect.bottom - windowrect.top)) / 2);
		}
	}
}


/*
===================
VID_NewWindow
===================
*/
void VID_NewWindow (int width, int height)
{
	viddef.width = width;
	viddef.height = height;

	cl.force_refdef = true;		// can't use a paused refdef
}


/*
============
VID_Init
============
*/
void VID_Init (void)
{
	static refimport_t	rimp;

	// Create the video variables so we know how to start the graphics drivers
	// this also gives us a nice single point of entry for the vid cvars rather than the previous mess
	vidvars.mode = Cvar_Get ("vid_mode", "3", CVAR_ARCHIVE);
	vidvars.xpos = Cvar_Get ("vid_xpos", "3", CVAR_ARCHIVE);
	vidvars.ypos = Cvar_Get ("vid_ypos", "22", CVAR_ARCHIVE);
	vidvars.fullscreen = Cvar_Get ("vid_fullscreen", "0", CVAR_ARCHIVE);
	vidvars.vsync = Cvar_Get ("gl_swapinterval", "0", CVAR_ARCHIVE);
	vidvars.width = Cvar_Get ("vid_width", "0", CVAR_ARCHIVE);
	vidvars.height = Cvar_Get ("vid_height", "0", CVAR_ARCHIVE);

	// Add some console commands that we want to handle
	Cmd_AddCommand ("vid_restart", VID_Restart_f);
	Cmd_AddCommand ("vid_centerwindow", VID_CenterWindow_f);

	// Start the graphics mode and load refresh
	cl.force_refdef = true;		// can't use a paused refdef
	S_StopAllSounds ();

	// refresh is now loaded for the first time
	cl.refresh_prepped = false;
	cls.disable_screen = true;

	if (reflib_active && re && re->Shutdown)
		re->Shutdown ();

	Com_Printf ("--------- Loading refresh ---------\n");

	rimp.vars = &vidvars;
	rimp.Cmd_AddCommand = Cmd_AddCommand;
	rimp.Cmd_RemoveCommand = Cmd_RemoveCommand;
	rimp.Cmd_Argc = Cmd_Argc;
	rimp.Cmd_Argv = Cmd_Argv;
	rimp.Cmd_ExecuteText = Cbuf_ExecuteText;
	rimp.Con_Printf = VID_Printf;
	rimp.Scratch_Alloc = Scratch_Alloc;
	rimp.Scratch_AllocSize = Scratch_AllocSize;
	rimp.Scratch_FreeToLowMark = Scratch_FreeToLowMark;
	rimp.Scratch_LowMark = Scratch_LowMark;
	rimp.Hunk_Begin = Hunk_Begin;
	rimp.Hunk_Alloc = Hunk_Alloc;
	rimp.Hunk_End = Hunk_End;
	rimp.Hunk_Free = Hunk_Free;
	rimp.Sys_Error = VID_Error;
	rimp.FS_LoadFile = FS_LoadFile;
	rimp.FS_FreeFile = FS_FreeFile;
	rimp.FS_Gamedir = FS_Gamedir;
	rimp.Cvar_Get = Cvar_Get;
	rimp.Cvar_Set = Cvar_Set;
	rimp.Cvar_SetValue = Cvar_SetValue;
	rimp.Vid_NewWindow = VID_NewWindow;
	rimp.PrepVideoMenu = M_PrepVideoMenu;
	rimp.SendKeyEvents = Sys_SendKeyEvents;

	// set up the refresh api and init d3d
	re = GetRefAPI (&rimp);
	re->vars = &vidvars;

	if (re->api_version != API_VERSION)
	{
		memset (&re, 0, sizeof (re));
		reflib_active = false;
		Com_Error (ERR_FATAL, "refresh has incompatible api_version");
	}

	// now set up the window before we init the renderer
	WNDCLASS wc;

	// Register the frame class
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = GetModuleHandle (NULL);
	wc.hIcon = 0;
	wc.hCursor = LoadCursor (NULL, IDC_ARROW);
	wc.hbrBackground = (void *) COLOR_GRAYTEXT;
	wc.lpszMenuName = 0;
	wc.lpszClassName = WINDOW_CLASS_NAME;

	if (!RegisterClass (&wc))
		VID_Error (ERR_FATAL, "Couldn't register window class");

	if (re->Init () == -1)
	{
		re->Shutdown ();
		memset (&re, 0, sizeof (re));
		reflib_active = false;
		Com_Error (ERR_FATAL, "VID_LoadRefresh : failed");
	}

	Com_Printf ("-----------------------------------\n");
	reflib_active = true;
	cls.disable_screen = false;
}


/*
============
VID_Shutdown
============
*/
void VID_Shutdown (void)
{
	if (reflib_active)
	{
		re->Shutdown ();
		memset (&re, 0, sizeof (re));
		reflib_active = false;
	}
}


